from bs4 import BeautifulSoup
import requests
import re

CAC40_URL = "https://companiesmarketcap.com/cac-40/largest-companies-by-market-cap/"
AMERICAN_LARGEST_COMPANIES = (
    "https://companiesmarketcap.com/usa/largest-companies-in-the-usa-by-market-cap/"
)


def get_cac40_symbols() -> list[str]:
    soup = BeautifulSoup(requests.get(CAC40_URL).content, "lxml")
    raw_codes = soup.find_all("div", class_="company-code")
    return list(map(lambda input: str(re.sub("<[^>]*>", "", str(input))), raw_codes))


def get_largest_symbols_by_country_normal(n: int = 500) -> list[str]:
    n_per_page = 100
    assert (
        n >= n_per_page and n <= 3571
    ), "Error, provide a valid number of symbol, <= 3571 and >= 100."

    raw_codes = []
    pages = (n // n_per_page) + 1

    for i in range(pages):
        soup = BeautifulSoup(
            requests.get(f"{AMERICAN_LARGEST_COMPANIES}?page={i+1}").content, "lxml"
        )
        raw_codes.extend(soup.find_all("div", class_="company-code"))
    return list(map(lambda input: str(re.sub("<[^>]*>", "", str(input))), raw_codes))[
        :n
    ]



from datetime import datetime
import numpy as np
import pandas as pd
import numpy.typing as npt
from typing import Literal
import matplotlib.pyplot as plt
from tqdm import tqdm
from fitter import Fitter
from psutil import cpu_count


def generate_brownian_paths(
    n_paths: int,
    n_steps: int,
    T: float | int,
    mu: float | int,
    sigma: float | int,
    s0: float | int,
    brownian_type: Literal["ABM", "GBM"] = "ABM",
    get_time: bool = True,
) -> npt.NDArray[np.float64] | tuple[npt.NDArray[np.float64], npt.NDArray[np.float64]]:
    dt = T / n_steps
    paths = np.zeros((n_paths, n_steps + 1))
    paths[:, 0] = np.log(s0)

    for i in tqdm(range(0, n_paths)):
        for j in range(n_steps):
            paths[i, j + 1] = (
                paths[i, j]
                + (mu * 0.5 * sigma**2) * dt
                + sigma * np.random.normal(0, np.sqrt(dt))
            )

    if brownian_type == "GBM":
        paths = np.exp(paths)

    return np.linspace(0, T, n_steps + 1), paths if get_time is True else paths


def account_evolution(risk_free_rate: float, time: float) -> float:
    return np.exp(risk_free_rate * time)


def plot_paths_with_distribution(
    time: npt.NDArray[np.float64],
    paths: npt.NDArray[np.float64],
    title: str = "Simulated Brownian Motion",
) -> None:
    fig = plt.figure(figsize=(15, 4))
    fig.suptitle(title)
    ax1 = fig.add_subplot(121)
    ax2 = fig.add_subplot(122)
    for y in paths:
        ax1.plot(time, y)
    ax1.set_xlabel("Time")
    ax1.set_ylabel("Price")
    ax1.set_title(f"{paths.shape[0]} simulations")
    ax1.grid(True)
    ax2.hist(
        paths[:, -1],
        density=True,
        bins=75,
        facecolor="blue",
        alpha=0.3,
        label="Frequency of X(T)",
    )
    ax2.hist(
        paths[:, paths.shape[-1] // 4],
        density=True,
        bins=75,
        facecolor="red",
        alpha=0.3,
        label="Frequency of X(T/4)",
    )
    ax2.set_xlabel("Price")
    ax2.set_ylabel("Density")
    ax2.set_title("Distribution at X(T) and X(T/4)")
    ax2.legend()
    ax2.grid(True)


def generate_open(
    open_divided_by_close: pd.Series, generated_close: npt.NDArray[np.float64]
) -> npt.NDArray[np.float64]:
    f = Fitter(
        open_divided_by_close.values,
        distributions=["laplace"],
    )
    f.fit(max_workers=cpu_count(), n_jobs=cpu_count())
    return (
        np.random.laplace(
            f.fitted_param["laplace"][0],
            f.fitted_param["laplace"][-1],
            generated_close.shape[0],
        )
    ) * generated_close


def generate_high_coef(
    high_from_body: pd.Series, n: int = 1000
) -> npt.NDArray[np.float64]:
    f = Fitter(
        high_from_body.values,
        distributions=["expon"],
    )
    f.fit(max_workers=cpu_count(), n_jobs=cpu_count())

    return 1 + np.random.exponential(f.fitted_param["expon"][-1], n)


def generate_low_coef(
    low_from_body: pd.Series, n: int = 1000
) -> npt.NDArray[np.float64]:
    f = Fitter(
        low_from_body.values,
        distributions=["expon"],
    )
    f.fit(max_workers=cpu_count(), n_jobs=cpu_count())
    return np.random.exponential(f.fitted_param["expon"][-1], n)


def generate_brownian_candle_from_dataframe(
    dataframe: pd.DataFrame,
    timeframe: Literal[
        "1min",
        "2min",
        "5min",
        "15min",
        "30min",
        "1hour",
        "2hour",
        "4hour",
        "12hour",
        "1day",
    ] = "1day",
    generated_dataframe_length: int = 5000,
    date_index: bool = True,
) -> pd.DataFrame:
    """Generate a new dataframe with the same structure (global continuous distribution) as the initial dataframe but with generated data using a geometric brownian motion.

    Args:
    ----
        dataframe (pd.DataFrame): The initial dataframe, it should be a real dataframe containing OHLC data. I must contains OHLC columns. This dataframe will to extract the mean return, the std return, the initial price and the distribution of the open, high and low prices.

        timeframe (Literal[ &quot;1min&quot;, &quot;2min&quot;, &quot;5min&quot;, &quot;15min&quot;, &quot;30min&quot;, &quot;1hour&quot;, &quot;2hour&quot;, &quot;4hour&quot;, &quot;12hour&quot;, &quot;1day&quot;, ], optional): The timeframe or data interval frequency, it's used to compute the period parameter. Defaults to "1day".

        generated_dataframe_length (int, optional): The number of record to generate in the new dataframe. Defaults to 5000.

        date_index (bool, optional): Create as of today date index. Defaults to True.

    Returns:
    ----
        pd.DataFrame: The generated dataframe and it's candle stick.
    """
    assert timeframe in [
        "1min",
        "2min",
        "5min",
        "15min",
        "30min",
        "1hour",
        "2hour",
        "4hour",
        "12hour",
        "1day",
    ], "timeframe must be one of the following: 1min, 2min, 5min, 15min, 30min, 1hour, 2hour, 4hour, 12hour, 1day"
    assert isinstance(dataframe, pd.DataFrame), "dataframe must be a pandas DataFrame"
    assert set(dataframe.columns).issuperset(
        ["Open", "High", "Low", "Close"]
    ), "dataframe must contains Open, High, Low, Close columns"
    assert (
        generated_dataframe_length >= 10
    ), "generated_dataframe_length must be greater than 10"

    timeframe_annualized = {
        "1min": int(365 * 24 / 1 * 60),
        "2min": int(365 * 24 / 1 * 30),
        "5min": int(365 * 24 / 1 * 12),
        "15min": int(365 * 24 / 1 * 4),
        "30min": int(365 * 24 / 1 * 2),
        "1hour": int(365 * 24 / 1),
        "2hour": int(365 * 24 / 2),
        "4hour": int(365 * 24 / 4),
        "12hour": int(365 * 24 / 12),
        "1day": 365,
    }

    if not "Returns" in dataframe.columns:
        dataframe["Returns"] = dataframe.Close.pct_change().fillna(0)

    dataframe["High_from_body"] = dataframe.apply(
        lambda row: row["High"] / row["Close"]
        if row["Open"] <= row["Close"]
        else row["High"] / row["Open"],
        axis=1,
    )

    dataframe["Low_from_body"] = dataframe.apply(
        lambda row: 1 - (row["Low"] / row["Close"])
        if row["Open"] >= row["Close"]
        else 1 - (row["Low"] / row["Open"]),
        axis=1,
    )

    dataframe["Open_by_close"] = dataframe["Open"] / dataframe["Close"]

    time, sim = generate_brownian_paths(
        1,
        generated_dataframe_length - 1,
        generated_dataframe_length / timeframe_annualized.get(timeframe, 365),
        dataframe.Returns.mean() * timeframe_annualized.get(timeframe, 365),
        dataframe.Returns.std() * timeframe_annualized.get(timeframe, 365) ** 0.5,
        dataframe.Close.iloc[0],
        "GBM",
    )

    generated_close = sim[-1, :]
    generated_open = generate_open(dataframe["Open_by_close"], generated_close)

    generated_high_coefs = generate_high_coef(
        dataframe["High_from_body"], generated_dataframe_length
    )
    generated_low_coefs = generate_low_coef(
        dataframe["Low_from_body"], generated_dataframe_length
    )

    generated_ohlc = pd.DataFrame(
        {
            "Open": generated_open,
            "High_from_body": generated_high_coefs,
            "Low_from_body": generated_low_coefs,
            "Close": generated_close,
        }
    )

    generated_ohlc["High"] = generated_ohlc.apply(
        lambda row: row["High_from_body"] * row["Close"]
        if row["Open"] <= row["Close"]
        else row["High_from_body"] * row["Open"],
        axis=1,
    )

    generated_ohlc["Low"] = generated_ohlc.apply(
        lambda row: (1 - row["Low_from_body"]) * row["Close"]
        if row["Open"] >= row["Close"]
        else (1 - row["Low_from_body"]) * row["Open"],
        axis=1,
    )
    if date_index:
        timeframe_to_freq = {
            "1min": "1T",
            "2min": "2T",
            "5min": "5T",
            "15min": "15T",
            "30min": "30T",
            "1hour": "1H",
            "2hour": "2H",
            "4hour": "4H",
            "12hour": "12H",
            "1day": "1D",
        }

        generated_ohlc.set_index(
            pd.date_range(
                end=datetime.now(),
                periods=generated_dataframe_length,
                freq=timeframe_to_freq[timeframe],
                normalize=True,
                name="Date",
            ),
            inplace=True,
        )
    return generated_ohlc.drop(columns=["High_from_body", "Low_from_body"])



from datetime import datetime
from typing import Optional, Callable, Literal
import os
import pandas as pd
import json
import time
from random import randint
from kucoin.client import Market
from concurrent.futures import ThreadPoolExecutor, as_completed
from tqdm import tqdm


def build_multi_crypto_dataframe(
    symbols: set,
    drop_na: bool = False,
    column_to_keep: str = "Close",
    timeframe: Literal[
        "1min",
        "2min",
        "5min",
        "15min",
        "30min",
        "1hour",
        "2hour",
        "4hour",
        "12hour",
        "1day",
    ] = "1day",
) -> pd.DataFrame:
    """Build a multi cryptocurrencies dataframe from a set of symbols. This dataframe will contain the closing price of each symbol or any other column specified in `column_to_keep`. This function is useful to build a dataframe for a multi cryptocurrencies portfolio.

    Args:
    -----
        symbols (set): The name of the symbols to download.

        drop_na (bool, optional): Whether or not to drop NaNs. Defaults to False.

        column_to_keep (str, optional): The name of the column to keep. Defaults to "Close".

        timeframe (Literal[ &quot;1min&quot;, &quot;2min&quot;, &quot;5min&quot;, &quot;15min&quot;, &quot;30min&quot;, &quot;1hour&quot;, &quot;2hour&quot;, &quot;4hour&quot;, &quot;12hour&quot;, &quot;1day&quot;, ], optional): The data granularity. Defaults to "1day".

    Returns:
    -----
        pd.DataFrame: The multi cryptocurrencies dataframe.
    """
    df = pd.DataFrame()
    for symbol in tqdm(symbols, desc="Fetching symbols..."):
        currency_df = download_crypto_historical_data(
            symbol, timeframe, compute_return=False, refresh_list_of_symbol=False
        )
        currency_df.index = currency_df.index.map(
            lambda x: datetime(x.year, x.month, x.day)
        )
        if column_to_keep not in currency_df.columns:
            raise ValueError(
                f"Column {column_to_keep} not in {symbol} dataframe. Please choose another column."
            )
        df = pd.concat(
            [
                df,
                currency_df[[f"{column_to_keep}"]].rename(
                    columns={f"{column_to_keep}": f"{symbol}_{column_to_keep}"}
                ),
            ],
            axis=1,
        )
    return df if not drop_na else df.dropna()


def download_crypto_historical_data(
    symbol: str,
    timeframe: str = "1hour",
    compute_return: bool = True,
    refresh_list_of_symbol: bool = True,
) -> pd.DataFrame:
    """Fetch a cryptocurrency historical data from Kucoin for a given symbol and timeframe.

    Args:
    ----
        symbol (str): The symbol to download.

        timeframe (str, optional): The timeframe of the data to download. Defaults to "1hour".

        compute_return (bool, optional): Whether or not to compute the return on close T and T-1. Defaults to True.

        refresh_list_of_symbol (bool, optional): Refresh the list of symbols/ticker from kucoin, it must be true for the first execution of the code. Defaults to True.

    Returns:
    ----
        pd.DataFrame: The historical data of the symbol with date index
    """
    service = CryptoService()
    if refresh_list_of_symbol is True:
        service.refresh_list_of_symbols()  # Uncomment for the first usage
    df = service.get_history_of_symbol(f"{symbol}", timeframe)
    df["Date"] = df["Timestamp"].apply(datetime.fromtimestamp)
    df = df.set_index("Date")
    df = df.loc[~df.index.duplicated(), :]
    df = df.sort_index()

    timeframe_to_freq = {
        "1min": "1T",
        "2min": "2T",
        "5min": "5T",
        "15min": "15T",
        "30min": "30T",
        "1hour": "1H",
        "2hour": "2H",
        "4hour": "4H",
        "12hour": "12H",
        "1day": "1D",
    }
    if compute_return is True:
        df["Returns"] = df["Close"].pct_change()
    df = df.fillna(0.0)
    return df  # .asfreq(timeframe_to_freq[timeframe])#.ffill()


class CryptoService:
    class KucoinDataFetcher:
        __client = Market(url="https://api.kucoin.com")
        __timeframes_in_s: dict[str, int] = {
            "1min": 60,
            "2min": 120,
            "5min": 300,
            "15min": 900,
            "30min": 1800,
            "1hour": 3600,
            "2hour": 7200,
            "4hour": 14400,
            "12hour": 43200,
            "1day": 86400,
        }
        timeframes: tuple[str] = tuple(__timeframes_in_s.keys())

        def __construct_timestamp_list(
            self,
            start_timestamp: int,
            end_timestamp: int,
            timeframe: str,
            exchange_limit: int = 1500,
        ) -> list[int]:
            """Private function that generates a list of timestamps spaced of `exchange_limit` times `timeframe`.
            Args:
                start_timestamp (str): The initial timestamp.
                end_timestamp (str): The final timestamp.
                timeframe (str): The desired timeframe, sould be 1min, 2min, 5min, 15min, 1hour, 4hour, 1day...
                exchange_limit (int, optional): The exchange limit : 1500 for Kucoin here.. Defaults to 1500.
            Returns:
                list[int]: The list of timestamps.
            """
            remaining = (end_timestamp - start_timestamp) // self.__timeframes_in_s[
                timeframe
            ]

            timestamp_i = end_timestamp
            timestamps = [timestamp_i]

            while remaining > exchange_limit:
                timestamp_i = (
                    timestamp_i - self.__timeframes_in_s[timeframe] * exchange_limit
                )
                remaining = remaining - exchange_limit
                timestamps.append(timestamp_i)

            timestamps.append(start_timestamp)

            return sorted(timestamps, reverse=True)

        @staticmethod
        def __handle_429_error(func: Callable) -> Callable:
            """Static private decorator that handle error 429 response from Kucoin API.
            Args:
                func (Callable): The function to wrap.
            Returns:
                Callable: The wrapper.
            """

            def wrapper(*args, **kwargs):
                passed = False
                while passed == False:
                    try:
                        return func(*args, **kwargs)
                    except:
                        time.sleep(randint(10, 100) / 100)
                        pass

            return wrapper

        @__handle_429_error
        def __get_data(
            self, symbol: str, start_at: int, end_at: int, timeframe: str = "15min"
        ) -> pd.DataFrame:
            """Private function that uses Kucoin API to get the data for a specific symbol and timeframe.
            Args:
                symbol (str): The symbol for the data we want to extract. Defaults to "BTC-USDT".
                start_at (int): The starting timestamp and. Note that this function could only outputs 1500 records. If the timeframe and the timestamps don't satisfy it, it will return a dataframe with 1500 records from the starting timestamp.
                end_at (int): The ending timestamp.
                timeframe (str, optional): The timeframe, it must be 1min, 2min, 5min, 15min, 1hour, 4hour, 1day... Defaults to '15min'.
            Returns:
                Optional[pd.DataFrame]: The dataframe containing historical records.
            """
            klines = self.__client.get_kline(
                f"{symbol}", timeframe, startAt=start_at, endAt=end_at
            )
            df = pd.DataFrame(
                klines,
                columns=[
                    "Timestamp",
                    "Open",
                    "Close",
                    "High",
                    "Low",
                    "Amount",
                    "Volume",
                ],
                dtype=float,
            )
            df["Timestamp"] = df["Timestamp"].astype(int)

            return df

        def get_symbols(self) -> list[str]:
            """Get a list of all symbols in kucoin
            Returns:
                list[str]: The symbol's list.
            """
            tickers = self.__client.get_all_tickers()
            if tickers is not None:
                return [tik["symbol"] for tik in tickers["ticker"]]
            raise ValueError("Error, no symbols found.")

        def download_history(
            self, symbol: str, since: str, timeframe: str, jobs: int = -1
        ) -> pd.DataFrame:
            """Download a set of historical data and save it.
            Args:
                symbol (str): The symbol for the data we want to extract. Defaults to "BTC-USDT".
                since (str): The initial date in format : dd-mm-yyyy.
                timeframe (str): The timeframe, it must be 1min, 2min, 5min, 15min, 1hour, 4hour, 1day... Defaults to '15min'.
                jobs (int, optional): The number of thread to parallelize the code. Defaults to -1.
            Raises:
                ValueError: Error in using parallelism.
            Returns:
                pd.DataFrame: The dataframe containing historical records.
            """
            try:
                start_timestamp = int(datetime.strptime(since, "%d-%m-%Y").timestamp())
            except:
                raise ValueError(
                    "Error, wrong date format, provide something in this format: dd-mm-yyyy"
                )
            end_timestamp = int(datetime.now().timestamp())

            assert (
                start_timestamp < end_timestamp
            ), "Error, the starting timestamp must be less than ending timestamp"

            timestamps = self.__construct_timestamp_list(
                start_timestamp, end_timestamp, timeframe
            )

            df = pd.DataFrame(
                columns=[
                    "Timestamp",
                    "Open",
                    "Close",
                    "High",
                    "Low",
                    "Amount",
                    "Volume",
                ],
                dtype=float,
            )
            if jobs == -1 or jobs == 1:
                for i in range(len(timestamps) - 1):
                    df = pd.concat(
                        [
                            df,
                            self.__get_data(
                                symbol, timestamps[i + 1], timestamps[i], timeframe
                            ),
                        ]
                    )

            elif jobs > 1 and jobs <= 25:
                with ThreadPoolExecutor(
                    max_workers=(len(timestamps) if len(timestamps) <= 25 else jobs)
                ) as executor:
                    processes = [
                        executor.submit(
                            self.__get_data,
                            symbol,
                            timestamps[i + 1],
                            timestamps[i],
                            timeframe,
                        )
                        for i in range(len(timestamps) - 1)
                    ]

                for task in as_completed(processes):
                    df = pd.concat([df, task.result()])
            else:
                raise ValueError(
                    "Error, jobs must be between 25 and 2 to use parallelism or -1 and 1 to do it sequentially."
                )

            df = df.sort_values(by="Timestamp")
            df.drop_duplicates(inplace=True)
            return df

    __kucoin_fetcher = KucoinDataFetcher()
    __base_dir = "./data/"
    __absolute_start_date = "01-01-2017"

    def get_list_of_symbols(
        self, base_currency: Optional[str] = None, quote_currency: Optional[str] = None
    ) -> list[str]:
        """Return the list of symbol in the database's file.
        Args:
            base_currency (Optional[str]): Filter by base currency.
            quote_currency (Optional[str]): Filter by quote currency.
        Raises:
            ValueError: If there is an error.
        Returns:
            list[str]: The list of symbols.
        """
        symbols = self.__open_symbols_list()
        if base_currency == None and quote_currency == None:
            return symbols
        elif base_currency == None and quote_currency is not None:
            return list(
                filter(
                    lambda symbol: symbol.split("-")[-1] == quote_currency.upper(),
                    symbols,
                )
            )
        elif quote_currency == None and base_currency is not None:
            return list(
                filter(
                    lambda symbol: symbol.split("-")[0] == base_currency.upper(),
                    symbols,
                )
            )
        elif quote_currency is not None and base_currency is not None:
            return list(
                filter(
                    lambda symbol: symbol.split("-")[0] == base_currency.upper()
                    and symbol.split("-")[-1] == quote_currency.upper(),
                    symbols,
                )
            )
        raise ValueError("Error, wrong parameters.")

    def get_history_of_symbol(self, symbol: str, timeframe: str) -> pd.DataFrame:
        """Function that get the history of a symbol.
        Args:
            symbol (str): The crypto symbol file.
            timeframe (str): The history's timeframe.
        Returns:
            pd.DataFrame: Records corresponding to the history.
        """
        assert (
            timeframe in self.__kucoin_fetcher.timeframes
        ), f"Error, timeframe must be in {self.__kucoin_fetcher.timeframes}"

        assert (
            symbol in self.get_list_of_symbols()
        ), "Error, wrong symbol, provide something like 'BTC-USDT'."
        return self.__refresh_or_download(symbol, timeframe)

    def refresh_list_of_symbols(self) -> None:
        """Function that refreshes the database's crypto listing."""
        self.__init_directories()
        with open(f"{self.__base_dir}list_available/crypto_available.json", "w") as f:
            json.dump({"listing": self.__kucoin_fetcher.get_symbols()}, f)

    def check_file_exists(self, symbol: str, timeframe: str) -> bool:
        """Verify if the history has already been fetched
        Args:
            symbol (str): The crypto symbol file.
            timeframe (str): The history's timeframe.
        Returns:
            bool: Whether or not the history is present.
        """
        return os.path.exists(f"{self.__base_dir}{timeframe}/{symbol}.csv")

    def __open_symbols_list(self) -> list[str]:
        """Private function that opens, reads and returns the list of symbols from the database's file.
        Returns:
            list[str]: The list of symbols.
        """
        with open(f"{self.__base_dir}list_available/crypto_available.json", "r") as f:
            return json.load(f)["listing"]

    def __refresh_or_download(self, symbol: str, timeframe: str) -> pd.DataFrame:
        """Private function used to refresh or download the history of a symbol.
        Args:
            symbol (str): The crypto symbol file.
            timeframe (str): The history's timeframe.
        Returns:
            pd.DataFrame: The complete history refreshed or not.
        """
        if self.check_file_exists(symbol, timeframe):
            # print("History present -> Checking for refresh...")
            data = pd.read_csv(
                f"{self.__base_dir}{timeframe}/{symbol}.csv",
                sep=",",
                dtype=float,
            )
            last_timestamp = data["Timestamp"].iloc[-1]

            since = datetime.fromtimestamp(last_timestamp).strftime("%d-%m-%Y")
            # print(f"Last timestamp : {last_timestamp} = {since}")
            if since == datetime.now().strftime("%d-%m-%Y"):
                return data
            new_data = self.__kucoin_fetcher.download_history(
                symbol, since, timeframe, 16
            )

            data = pd.concat([data, new_data]).drop_duplicates(ignore_index=True)
            data["Timestamp"] = data["Timestamp"].astype(int)
            data.to_csv(
                f"{self.__base_dir}{timeframe}/{symbol}.csv", sep=",", index=False
            )
            # print("Finished")
            return data
        else:  # Download full history
            # print("No history -> Downloading full history...")
            data = self.__kucoin_fetcher.download_history(
                symbol, self.__absolute_start_date, timeframe, 16
            )
            data["Timestamp"] = data["Timestamp"].astype(int)
            data.to_csv(
                f"{self.__base_dir}{timeframe}/{symbol}.csv", sep=",", index=False
            )
            # print("Finished")
            return data

    def __init_directories(self) -> None:
        """Private function that init all directories."""
        for tf in self.__kucoin_fetcher.timeframes:
            os.makedirs(f"{self.__base_dir}{tf}", exist_ok=True)
        os.makedirs(f"{self.__base_dir}list_available", exist_ok=True)



from datetime import datetime
from scipy.stats import skew, kurtosis
from typing import Callable, Literal, Optional, Union, Tuple
import pandas as pd
import numpy as np
import plotly.graph_objects as go
import plotly.figure_factory as ff
from plotly.subplots import make_subplots
from tqdm import tqdm

from quant_invest_lab.metrics import (
    sharpe_ratio,
    calmar_ratio,
    sortino_ratio,
    max_drawdown,
    drawdown,
    kelly_criterion,
    value_at_risk,
    conditional_value_at_risk,
)


def ohlc_long_only_backtester(
    df: pd.DataFrame,
    long_entry_function: Callable[[pd.Series, pd.Series], bool],
    long_exit_function: Callable[[pd.Series, pd.Series, int], bool],
    timeframe: Literal[
        "1min",
        "2min",
        "5min",
        "15min",
        "30min",
        "1hour",
        "2hour",
        "4hour",
        "12hour",
        "1day",
    ],
    take_profit: float = np.inf,
    stop_loss: float = np.inf,
    initial_equity: int = 1000,
    maker_fees: Optional[float] = 0.001,
    taker_fees: Optional[float] = 0.001,
    get_trade_df: bool = False,
    get_returns_df: bool = False,
    parameter_optimization: bool = False,
    plot_result: bool = True,
) -> Union[None, Union[float, Union[pd.DataFrame, Tuple[pd.DataFrame, pd.DataFrame]]]]:
    """Run a backtest with long only position on a OHLC dataset.

    Args:
    ----
        df (pd.DataFrame): The dataframe containing the OHLC data.

        long_entry_function (Callable[[pd.Series, pd.Series], bool]): The long entry function, it should take 2 arguments, the current row and the previous row and return True or False depending on your strategy.

        long_exit_function (Callable[[pd.Series, pd.Series, int], bool]): The long exit function, it should take 3 arguments, the current row, the previous row and the number of timeframe count since the last entry order, and return True or False depending on your strategy.

        timeframe (Literal[ &quot;1min&quot;, &quot;2min&quot;, &quot;5min&quot;, &quot;15min&quot;, &quot;30min&quot;, &quot;1hour&quot;, &quot;2hour&quot;, &quot;4hour&quot;, &quot;12hour&quot;, &quot;1day&quot;, ]): The timeframe granularity of the dataframe.

        take_profit (float, optional): The percent of the buy price to add to create a stop order and take the profit associated. Defaults to np.inf.

        stop_loss (float, optional): The percent of the buy price to cut to create a stop order and stop the loss associated. Defaults to np.inf.

        initial_equity (int, optional): The initial capital. Defaults to 1000.

        maker_fees (Optional[float], optional): The fees applied, here maker for limit orders (not yet implemented). Defaults to 0.001.

        taker_fees (Optional[float], optional): The fees applied, here taker for spot orders. Defaults to 0.001.

        get_trade_df (bool, optional): Whether or not to return the trade dataframe (summary of trades) . Defaults to False.

        get_returns_df (bool, optional): Whether or not to return the strategy returns dataframe. Defaults to False.

        parameter_optimization (bool, optional): This parameter is useful when running fitting/optimization it prints nothing but the final strategy return. Defaults to False.

        plot_result (bool, optional): Plot equity, price, drawdown, distribution of the strategy. Defaults to True.

    Returns:
    -----
        Union[None, Union[float, Union[pd.DataFrame, Tuple[pd.DataFrame, pd.DataFrame]]]]: Nothing or the strategy total return or the trade_df or the return_df or the trade_df and the return_df.
    """
    assert timeframe in [
        "1min",
        "2min",
        "5min",
        "15min",
        "30min",
        "1hour",
        "2hour",
        "4hour",
        "12hour",
        "1day",
    ], "timeframe must be one of the following: 1min, 2min, 5min, 15min, 30min, 1hour, 2hour, 4hour, 12hour, 1day"
    assert df.shape[0] > 1, "Dataframe must have at least 2 rows"
    assert set({"Open", "High", "Low", "Close", "Returns"}).issubset(
        df.columns
    ), "Dataframe must have columns Open, High, Low, Close, Returns"
    timeframe_annualized = {
        "1min": int(365 * 24 / 1 * 60),
        "2min": int(365 * 24 / 1 * 30),
        "5min": int(365 * 24 / 1 * 12),
        "15min": int(365 * 24 / 1 * 4),
        "30min": int(365 * 24 / 1 * 2),
        "1hour": int(365 * 24 / 1),
        "2hour": int(365 * 24 / 2),
        "4hour": int(365 * 24 / 4),
        "12hour": int(365 * 24 / 12),
        "1day": 365,
    }
    N = timeframe_annualized[timeframe]
    ohlcv_df = df.copy()
    previous_row = ohlcv_df.iloc[0]
    position_opened = False
    timeframe_count = 0
    current_trade: dict[str, float | datetime | int | str | pd.Timestamp] = {}

    trades_df = pd.DataFrame(
        columns=[
            "entry_date",
            "entry_price",
            "entry_reason",
            "exit_date",
            "exit_price",
            "exit_reason",
            "trade_return",
        ]
    )

    returns_df = pd.DataFrame(columns=["Returns"])

    for index, row in tqdm(
        ohlcv_df[1:].iterrows(),
        desc="Backtesting...",
        total=ohlcv_df.shape[0] - 1,
        leave=False,
    ):
        if position_opened is False and long_entry_function(row, previous_row) is True:
            position_opened = True
            current_trade["entry_date"] = index
            current_trade["entry_price"] = row.Close
            current_trade["entry_reason"] = "Entry position triggered"
            timeframe_count = 1
            entry_price = row.Close
        elif (
            position_opened is True
            and current_trade["entry_price"] * (1 + take_profit) <= row.High
        ):
            position_opened = False
            current_trade["exit_date"] = index
            current_trade["exit_price"] = current_trade["entry_price"] * (
                1 + take_profit
            )
            current_trade["exit_reason"] = "Take profit triggered"

            rets = ohlcv_df.loc[
                current_trade["entry_date"] : current_trade["exit_date"]
            ].Returns
            if isinstance(taker_fees, float) and taker_fees > 0:
                rets.iloc[0] = rets.iloc[0] - taker_fees
                rets.iloc[-1] = rets.iloc[-1] - taker_fees
            current_trade["trade_return"] = ((rets + 1).cumprod().iloc[-1]) - 1  # ret

            trades_df = pd.concat(
                [trades_df, pd.DataFrame([current_trade])], ignore_index=True
            )
            returns_df = pd.concat(
                [returns_df, pd.DataFrame({"Returns": rets.values}, index=rets.index)],
                ignore_index=False,
            )
            timeframe_count = 0

            current_trade = {}
        elif (
            position_opened is True
            and current_trade["entry_price"] * (1 - stop_loss) >= row.Low
        ):
            position_opened = False
            current_trade["exit_date"] = index
            current_trade["exit_price"] = float(current_trade["entry_price"]) * (
                1 - stop_loss
            )
            current_trade["exit_reason"] = "Stop loss triggered"

            rets = ohlcv_df.loc[
                current_trade["entry_date"] : current_trade["exit_date"]
            ].Returns
            if isinstance(taker_fees, float) and taker_fees > 0:
                rets.iloc[0] = rets.iloc[0] - taker_fees
                rets.iloc[-1] = rets.iloc[-1] - taker_fees
            current_trade["trade_return"] = ((rets + 1).cumprod().iloc[-1]) - 1  # ret

            trades_df = pd.concat(
                [trades_df, pd.DataFrame([current_trade])], ignore_index=True
            )
            returns_df = pd.concat(
                [returns_df, pd.DataFrame({"Returns": rets.values}, index=rets.index)],
                ignore_index=False,
            )
            timeframe_count = 0
            current_trade = {}
        elif (
            position_opened is True
            and long_exit_function(row, previous_row, timeframe_count) is True
        ):
            position_opened = False
            current_trade["exit_date"] = index
            current_trade["exit_price"] = row.Close
            current_trade["exit_reason"] = "Exit position triggered"

            rets = ohlcv_df.loc[
                current_trade["entry_date"] : current_trade["exit_date"]
            ].Returns

            if isinstance(taker_fees, float) and taker_fees > 0:
                rets.iloc[0] = rets.iloc[0] - taker_fees
                rets.iloc[-1] = rets.iloc[-1] - taker_fees

            current_trade["trade_return"] = ((rets + 1).cumprod().iloc[-1]) - 1  # ret

            trades_df = pd.concat(
                [trades_df, pd.DataFrame([current_trade])], ignore_index=True
            )
            returns_df = pd.concat(
                [returns_df, pd.DataFrame({"Returns": rets.values}, index=rets.index)],
                ignore_index=False,
            )
            timeframe_count = 0
            current_trade = {}
        else:
            timeframe_count += 1
        previous_row = row

    returns_df["Cum_Returns"] = (returns_df["Returns"] + 1).cumprod()
    returns_df["Drawdown"] = drawdown(returns_df["Returns"])
    ohlcv_df["Cum_Returns"] = (df["Returns"] + 1).cumprod()
    ohlcv_df["Drawdown"] = drawdown(df["Returns"])
    if parameter_optimization is True:
        if len(trades_df) > 0:
            return returns_df["Cum_Returns"].iloc[-1]
        return 0.0

    assert len(trades_df) > 0, "No trades were generated"
    trades_df["trade_duration"] = trades_df["exit_date"] - trades_df["entry_date"]

    good_trades = trades_df.loc[trades_df["trade_return"] > 0]
    bad_trades = trades_df.loc[trades_df["trade_return"] < 0]
    total_trades = len(trades_df)

    print(f"{'  Initial informations  ':-^50}")
    print(f"Period: [{str(ohlcv_df.index[0])}] -> [{str(ohlcv_df.index[-1])}]")
    print(f"Intial balance: {initial_equity:.2f} $")
    if taker_fees is not None and maker_fees is not None:
        print(
            f"Taker fees: {taker_fees*100:.2f} %, Maker fees: {maker_fees*100:.2f} %, All the metrics will be calculated considering these fees"
        )
    else:
        print("No fees considered here.")

    print(f"\n{'  Strategy performances  ':-^50}")

    print(
        f'Strategy final net balance: {returns_df["Cum_Returns"].iloc[-1]*initial_equity:.2f} $, return: {(returns_df["Cum_Returns"].iloc[-1]-1)*100:.2f} %'
    )
    print(
        f'Buy & Hold final net balance: {ohlcv_df["Cum_Returns"].iloc[-1]*initial_equity:.2f} $, returns: {(ohlcv_df["Cum_Returns"].iloc[-1]-1)*100:.2f} %'
    )
    print(f"Strategy winrate ratio: {100 * len(good_trades) / total_trades:.2f} %")
    print(
        f"Strategy profit factor ratio: {abs(good_trades['trade_return'].mean()/bad_trades['trade_return'].mean()):.2f}"
    )

    print(f"\n{'  Returns statistical information  ':-^50}")

    print(
        f"Expected return : {100*returns_df['Returns'].mean():.2f} %, annuzalized: {100*returns_df['Returns'].mean()*N:.2f} %"
    )
    print(
        f"Median return : {100*returns_df['Returns'].median():.2f} %, annuzalized: {100*returns_df['Returns'].median()*N:.2f} %"
    )
    print(
        f'Expected volatility: {100*returns_df["Returns"].std():.2f} %, annualized: {100*returns_df["Returns"].std()*(N**0.5):.2f} %'
    )
    print(
        f'Skewness: {skew(returns_df["Returns"].values):.2f}, <0 = left tail, >0 = right tail -> the higher the better'
    )
    print(
        f'Kurtosis: {kurtosis(returns_df["Returns"].values):.2f}',
        ", >3 = fat tails, <3 = thin tails -> the lower the better",
    )
    print(f"{timeframe}-95%-VaR: {100*value_at_risk(returns_df['Returns']):.2f} %")
    print(
        f"{timeframe}-95%-CVaR: {100*conditional_value_at_risk(returns_df['Returns']):.2f} %"
    )

    print(f"\n{'  Strategy statistical information  ':-^50}")
    print(f"Max drawdown: {100*max_drawdown(returns_df['Returns']):.2f} %")
    print(f"Kelly criterion: {100*kelly_criterion(returns_df.Returns):.2f} %")
    print(
        f"Sharpe ratio (annualized): {sharpe_ratio(returns_df['Returns'], N,risk_free_rate=N*ohlcv_df.Returns.mean()):.2f} (risk free rate = buy and hold)"
    )
    print(
        f"Sortino ratio (annualized): {sortino_ratio(returns_df['Returns'], N,risk_free_rate=N*ohlcv_df.Returns.mean()):.2f} (risk free rate = buy and hold)"
    )
    print(
        f"Calmar ratio (annualized): {calmar_ratio(returns_df['Returns'], N):.2f} (risk free rate = buy and hold)"
    )

    print(f"\n{'  Trades informations  ':-^50}")
    print(f"Mean trade return : {100*trades_df['trade_return'].mean():.2f} %")
    print(f"Median trade return : {100*trades_df['trade_return'].median():.2f} %")
    print(f'Mean trade volatility: {100*trades_df["trade_return"].std():.2f} %')
    print(
        f"Mean trade duration: {str((trades_df['trade_duration']).mean()).split('.')[0]}"
    )

    print(f"Total trades: {total_trades}")

    print(f"\n  Total good trades: {len(good_trades)}")
    print(f"  Mean good trades return: {100*good_trades['trade_return'].mean():.2f} %")
    print(
        f"  Median good trades return: {100*good_trades['trade_return'].median():.2f} %"
    )
    print(
        f"  Best trades return: {100*trades_df['trade_return'].max():.2f} % | Date: {trades_df.iloc[trades_df['trade_return'].idxmax()]['exit_date']} | Duration: {trades_df.iloc[trades_df['trade_return'].idxmax()]['trade_duration']}"
    )
    print(
        f"  Mean good trade duration: {str((good_trades['trade_duration']).mean()).split('.')[0]}"
    )
    print(f"\n  Total bad trades: {len(bad_trades)}")
    print(f"  Mean bad trades return: {100*bad_trades['trade_return'].mean():.2f} %")
    print(
        f"  Median bad trades return: {100*bad_trades['trade_return'].median():.2f} %"
    )
    print(
        f"  Worst trades return: {100*trades_df['trade_return'].min():.2f} % | Date: {trades_df.iloc[trades_df['trade_return'].idxmin()]['exit_date']} | Duration: {trades_df.iloc[trades_df['trade_return'].idxmin()]['trade_duration']}"
    )
    print(
        f"  Mean bad trade duration: {str((bad_trades['trade_duration']).mean()).split('.')[0]}"
    )

    print(f"\nExit reasons repartition: ")
    for reason, val in zip(
        trades_df.exit_reason.value_counts().index, trades_df.exit_reason.value_counts()
    ):
        print(f"- {reason}: {val}")
    if plot_result is True:
        plot_from_trade_df(trades_df, ohlcv_df, returns_df)
    if get_returns_df and get_trade_df:
        return trades_df, returns_df
    if get_trade_df:
        return trades_df
    if get_returns_df:
        return returns_df


def ohlc_short_only_backtester(
    df: pd.DataFrame,
    short_entry_function: Callable[[pd.Series, pd.Series], bool],
    short_exit_function: Callable[[pd.Series, pd.Series, int], bool],
    timeframe: Literal[
        "1min",
        "2min",
        "5min",
        "15min",
        "30min",
        "1hour",
        "2hour",
        "4hour",
        "12hour",
        "1day",
    ],
    take_profit: float = np.inf,
    stop_loss: float = np.inf,
    initial_equity: int = 1000,
    maker_fees: Optional[float] = 0.001,
    taker_fees: Optional[float] = 0.001,
    get_trade_df: bool = False,
    get_returns_df: bool = False,
    parameter_optimization: bool = False,
    plot_result: bool = True,
) -> Union[None, Union[float, Union[pd.DataFrame, Tuple[pd.DataFrame, pd.DataFrame]]]]:
    """Run a backtest with short only position on a OHLC dataset.

    Args:
    ----
        df (pd.DataFrame): The dataframe containing the OHLC data.

        short_entry_function (Callable[[pd.Series, pd.Series], bool]): The short entry function, it should take 2 arguments, the current row and the previous row and return True or False depending on your strategy.

        short_exit_function (Callable[[pd.Series, pd.Series, int], bool]): The short exit function, it should take 3 arguments, the current row, the previous row and the number of timeframe count since the last entry order, and return True or False depending on your strategy.

        timeframe (Literal[ &quot;1min&quot;, &quot;2min&quot;, &quot;5min&quot;, &quot;15min&quot;, &quot;30min&quot;, &quot;1hour&quot;, &quot;2hour&quot;, &quot;4hour&quot;, &quot;12hour&quot;, &quot;1day&quot;, ]): The timeframe granularity of the dataframe.

        take_profit (float, optional): The percent of the buy price to add to create a stop order and take the profit associated. Defaults to np.inf.

        stop_loss (float, optional): The percent of the buy price to cut to create a stop order and stop the loss associated. Defaults to np.inf.

        initial_equity (int, optional): The initial capital. Defaults to 1000.

        maker_fees (Optional[float], optional): The fees applied, here maker for limit orders (not yet implemented). Defaults to 0.001.

        taker_fees (Optional[float], optional): The fees applied, here taker for spot orders. Defaults to 0.001.

        get_trade_df (bool, optional): Whether or not to return the trade dataframe (summary of trades) . Defaults to False.

        get_returns_df (bool, optional): Whether or not to return the strategy returns dataframe. Defaults to False.

        parameter_optimization (bool, optional): This parameter is useful when running fitting/optimization it prints nothing but the final strategy return. Defaults to False.

        plot_result (bool, optional): Plot equity, price, drawdown, distribution of the strategy. Defaults to True.

    Returns:
    -----
        Union[None, Union[float, Union[pd.DataFrame, Tuple[pd.DataFrame, pd.DataFrame]]]]: Nothing or the strategy total return or the trade_df or the return_df or the trade_df and the return_df.
    """
    assert timeframe in [
        "1min",
        "2min",
        "5min",
        "15min",
        "30min",
        "1hour",
        "2hour",
        "4hour",
        "12hour",
        "1day",
    ], "timeframe must be one of the following: 1min, 2min, 5min, 15min, 30min, 1hour, 2hour, 4hour, 12hour, 1day"
    assert df.shape[0] > 1, "Dataframe must have at least 2 rows"
    assert set({"Open", "High", "Low", "Close", "Returns"}).issubset(
        df.columns
    ), "Dataframe must have columns Open, High, Low, Close, Returns"
    timeframe_annualized = {
        "1min": int(365 * 24 / 1 * 60),
        "2min": int(365 * 24 / 1 * 30),
        "5min": int(365 * 24 / 1 * 12),
        "15min": int(365 * 24 / 1 * 4),
        "30min": int(365 * 24 / 1 * 2),
        "1hour": int(365 * 24 / 1),
        "2hour": int(365 * 24 / 2),
        "4hour": int(365 * 24 / 4),
        "12hour": int(365 * 24 / 12),
        "1day": 365,
    }
    N = timeframe_annualized[timeframe]
    ohlcv_df = df.copy()
    previous_row = ohlcv_df.iloc[0]
    position_opened = False
    timeframe_count = 0
    current_trade: dict[str, float | datetime | int | str | pd.Timestamp] = {}

    trades_df = pd.DataFrame(
        columns=[
            "entry_date",
            "entry_price",
            "entry_reason",
            "exit_date",
            "exit_price",
            "exit_reason",
            "trade_return",
        ]
    )

    returns_df = pd.DataFrame(columns=["Returns"])

    for index, row in tqdm(
        ohlcv_df[1:].iterrows(),
        desc="Backtesting...",
        total=ohlcv_df.shape[0] - 1,
        leave=False,
    ):
        if position_opened is False and short_entry_function(row, previous_row) is True:
            position_opened = True
            current_trade["entry_date"] = index
            current_trade["entry_price"] = row.Close
            current_trade["entry_reason"] = "Entry position triggered"
            timeframe_count = 1
            entry_price = row.Close
        elif (
            position_opened is True
            and current_trade["entry_price"] * (1 + stop_loss) <= row.High
        ):
            position_opened = False
            current_trade["exit_date"] = index
            current_trade["exit_price"] = current_trade["entry_price"] * (1 + stop_loss)
            current_trade["exit_reason"] = "Stop loss triggered"

            rets = -ohlcv_df.loc[
                current_trade["entry_date"] : current_trade["exit_date"]
            ].Returns
            if isinstance(taker_fees, float) and taker_fees > 0:
                rets.iloc[0] = rets.iloc[0] - taker_fees
                rets.iloc[-1] = rets.iloc[-1] - taker_fees
            current_trade["trade_return"] = ((rets + 1).cumprod().iloc[-1]) - 1  # ret

            trades_df = pd.concat(
                [trades_df, pd.DataFrame([current_trade])], ignore_index=True
            )
            returns_df = pd.concat(
                [returns_df, pd.DataFrame({"Returns": rets.values}, index=rets.index)],
                ignore_index=False,
            )
            timeframe_count = 0

            current_trade = {}
        elif (
            position_opened is True
            and current_trade["entry_price"] * (1 - take_profit) >= row.Low
        ):
            position_opened = False
            current_trade["exit_date"] = index
            current_trade["exit_price"] = float(current_trade["entry_price"]) * (
                1 - take_profit
            )
            current_trade["exit_reason"] = "Take profit triggered"

            rets = -ohlcv_df.loc[
                current_trade["entry_date"] : current_trade["exit_date"]
            ].Returns
            if isinstance(taker_fees, float) and taker_fees > 0:
                rets.iloc[0] = rets.iloc[0] - taker_fees
                rets.iloc[-1] = rets.iloc[-1] - taker_fees
            current_trade["trade_return"] = ((rets + 1).cumprod().iloc[-1]) - 1  # ret

            trades_df = pd.concat(
                [trades_df, pd.DataFrame([current_trade])], ignore_index=True
            )
            returns_df = pd.concat(
                [returns_df, pd.DataFrame({"Returns": rets.values}, index=rets.index)],
                ignore_index=False,
            )
            timeframe_count = 0
            current_trade = {}
        elif (
            position_opened is True
            and short_exit_function(row, previous_row, timeframe_count) is True
        ):
            position_opened = False
            current_trade["exit_date"] = index
            current_trade["exit_price"] = row.Close
            current_trade["exit_reason"] = "Exit position triggered"

            rets = -ohlcv_df.loc[
                current_trade["entry_date"] : current_trade["exit_date"]
            ].Returns

            if isinstance(taker_fees, float) and taker_fees > 0:
                rets.iloc[0] = rets.iloc[0] - taker_fees
                rets.iloc[-1] = rets.iloc[-1] - taker_fees

            current_trade["trade_return"] = ((rets + 1).cumprod().iloc[-1]) - 1  # ret

            trades_df = pd.concat(
                [trades_df, pd.DataFrame([current_trade])], ignore_index=True
            )
            returns_df = pd.concat(
                [returns_df, pd.DataFrame({"Returns": rets.values}, index=rets.index)],
                ignore_index=False,
            )
            timeframe_count = 0
            current_trade = {}
        else:
            timeframe_count += 1
        previous_row = row

    returns_df["Cum_Returns"] = (returns_df["Returns"] + 1).cumprod()
    returns_df["Drawdown"] = drawdown(returns_df["Returns"])
    ohlcv_df["Cum_Returns"] = (df["Returns"] + 1).cumprod()
    ohlcv_df["Drawdown"] = drawdown(df["Returns"])
    if parameter_optimization is True:
        if len(trades_df) > 0:
            return returns_df["Cum_Returns"].iloc[-1]
        return 0.0

    assert len(trades_df) > 0, "No trades were generated"
    trades_df["trade_duration"] = trades_df["exit_date"] - trades_df["entry_date"]

    good_trades = trades_df.loc[trades_df["trade_return"] > 0]
    bad_trades = trades_df.loc[trades_df["trade_return"] < 0]
    total_trades = len(trades_df)

    print(f"{'  Initial informations  ':-^50}")
    print(f"Period: [{str(ohlcv_df.index[0])}] -> [{str(ohlcv_df.index[-1])}]")
    print(f"Intial balance: {initial_equity:.2f} $")
    if taker_fees is not None and maker_fees is not None:
        print(
            f"Taker fees: {taker_fees*100:.2f} %, Maker fees: {maker_fees*100:.2f} %, All the metrics will be calculated considering these fees"
        )
    else:
        print("No fees considered here.")

    print(f"\n{'  Strategy performances  ':-^50}")

    print(
        f'Strategy final net balance: {returns_df["Cum_Returns"].iloc[-1]*initial_equity:.2f} $, return: {(returns_df["Cum_Returns"].iloc[-1]-1)*100:.2f} %'
    )
    print(
        f'Buy & Hold final net balance: {ohlcv_df["Cum_Returns"].iloc[-1]*initial_equity:.2f} $, returns: {(ohlcv_df["Cum_Returns"].iloc[-1]-1)*100:.2f} %'
    )
    print(f"Strategy winrate ratio: {100 * len(good_trades) / total_trades:.2f} %")
    print(
        f"Strategy profit factor ratio: {abs(good_trades['trade_return'].mean()/bad_trades['trade_return'].mean()):.2f}"
    )

    print(f"\n{'  Returns statistical information  ':-^50}")

    print(
        f"Expected return : {100*returns_df['Returns'].mean():.2f} %, annuzalized: {100*returns_df['Returns'].mean()*N:.2f} %"
    )
    print(
        f"Median return : {100*returns_df['Returns'].median():.2f} %, annuzalized: {100*returns_df['Returns'].median()*N:.2f} %"
    )
    print(
        f'Expected volatility: {100*returns_df["Returns"].std():.2f} %, annualized: {100*returns_df["Returns"].std()*(N**0.5):.2f} %'
    )
    print(
        f'Skewness: {skew(returns_df["Returns"].values):.2f}, <0 = left tail, >0 = right tail -> the higher the better'
    )
    print(
        f'Kurtosis: {kurtosis(returns_df["Returns"].values):.2f}',
        ", >3 = fat tails, <3 = thin tails -> the lower the better",
    )
    print(f"{timeframe}-95%-VaR: {100*value_at_risk(returns_df['Returns']):.2f} %")
    print(
        f"{timeframe}-95%-CVaR: {100*conditional_value_at_risk(returns_df['Returns']):.2f} %"
    )

    print(f"\n{'  Strategy statistical information  ':-^50}")
    print(f"Max drawdown: {100*max_drawdown(returns_df['Returns']):.2f} %")
    print(f"Kelly criterion: {100*kelly_criterion(returns_df.Returns):.2f} %")
    print(
        f"Sharpe ratio (annualized): {sharpe_ratio(returns_df['Returns'], N,risk_free_rate=N*ohlcv_df.Returns.mean()):.2f} (risk free rate = buy and hold)"
    )
    print(
        f"Sortino ratio (annualized): {sortino_ratio(returns_df['Returns'], N,risk_free_rate=N*ohlcv_df.Returns.mean()):.2f} (risk free rate = buy and hold)"
    )
    print(
        f"Calmar ratio (annualized): {calmar_ratio(returns_df['Returns'], N):.2f} (risk free rate = buy and hold)"
    )

    print(f"\n{'  Trades informations  ':-^50}")
    print(f"Mean trade return : {100*trades_df['trade_return'].mean():.2f} %")
    print(f"Median trade return : {100*trades_df['trade_return'].median():.2f} %")
    print(f'Mean trade volatility: {100*trades_df["trade_return"].std():.2f} %')
    print(
        f"Mean trade duration: {str((trades_df['trade_duration']).mean()).split('.')[0]}"
    )

    print(f"Total trades: {total_trades}")

    print(f"\n  Total good trades: {len(good_trades)}")
    print(f"  Mean good trades return: {100*good_trades['trade_return'].mean():.2f} %")
    print(
        f"  Median good trades return: {100*good_trades['trade_return'].median():.2f} %"
    )
    print(
        f"  Best trades return: {100*trades_df['trade_return'].max():.2f} % | Date: {trades_df.iloc[trades_df['trade_return'].idxmax()]['exit_date']} | Duration: {trades_df.iloc[trades_df['trade_return'].idxmax()]['trade_duration']}"
    )
    print(
        f"  Mean good trade duration: {str((good_trades['trade_duration']).mean()).split('.')[0]}"
    )
    print(f"\n  Total bad trades: {len(bad_trades)}")
    print(f"  Mean bad trades return: {100*bad_trades['trade_return'].mean():.2f} %")
    print(
        f"  Median bad trades return: {100*bad_trades['trade_return'].median():.2f} %"
    )
    print(
        f"  Worst trades return: {100*trades_df['trade_return'].min():.2f} % | Date: {trades_df.iloc[trades_df['trade_return'].idxmin()]['exit_date']} | Duration: {trades_df.iloc[trades_df['trade_return'].idxmin()]['trade_duration']}"
    )
    print(
        f"  Mean bad trade duration: {str((bad_trades['trade_duration']).mean()).split('.')[0]}"
    )

    print(f"\nExit reasons repartition: ")
    for reason, val in zip(
        trades_df.exit_reason.value_counts().index, trades_df.exit_reason.value_counts()
    ):
        print(f"- {reason}: {val}")
    if plot_result is True:
        plot_from_trade_df(trades_df, ohlcv_df, returns_df)
    if get_returns_df and get_trade_df:
        return trades_df, returns_df
    if get_trade_df:
        return trades_df
    if get_returns_df:
        return returns_df


def plot_from_trade_df(
    trade_df: pd.DataFrame, price_df: pd.DataFrame, returns_df: pd.DataFrame
) -> None:
    """Plot historical price, equity progression, drawdown evolution and return distribution.

    Args:
    ----
        trade_df (pd.DataFrame): The trade summary dataframe.
        price_df (pd.DataFrame): The historical price dataframe.
        returns_df (pd.DataFrame): The detailed strategy returns dataframe.
    """
    fig = make_subplots(
        rows=4,
        cols=1,
        subplot_titles=(
            "Historical price",
            "Equity progression",
            "Drawdown Evolution",
            "Return distribution",
        ),
        shared_xaxes=True,
    )

    fig.add_trace(
        go.Candlestick(
            name="Historical price",
            x=price_df.index,
            open=price_df["Open"],
            high=price_df["High"],
            low=price_df["Low"],
            close=price_df["Close"],
        ),
        row=1,
        col=1,
    )
    fig.update_yaxes(title_text="Historical price (USDT)", row=1, col=1)
    fig.update_xaxes(title_text="Datetime", row=1, col=1)
    trades = trade_df.copy()
    trades["date"] = trades["exit_date"]
    trades = trades.set_index("date")

    fig.add_trace(
        go.Scatter(
            name="Buy and hold cumulative return",
            x=price_df.index,
            y=price_df["Cum_Returns"],
            line={"shape": "hv", "color": "violet"},
        ),
        row=2,
        col=1,
    )
    fig.add_trace(
        go.Scatter(
            name="Strategy cumulative return",
            x=returns_df.index,
            y=returns_df["Cum_Returns"],
            line={"shape": "hv", "color": "salmon"},
        ),
        row=2,
        col=1,
    )
    fig.update_yaxes(type="log", row=2, col=1)
    fig.update_yaxes(title_text="Log cumulative returns", row=2, col=1)
    fig.update_xaxes(title_text="Datetime", row=2, col=1)
    fig.add_trace(
        go.Scatter(
            name="Buy and hold drawdown",
            x=price_df.index,
            y=price_df["Drawdown"],
            line={"shape": "hv", "color": "violet"},
            fill="tozeroy",
            fillcolor="rgba(238, 130, 238, 0.35)",
        ),
        row=3,
        col=1,
    )
    fig.add_trace(
        go.Scatter(
            name="Strategy drawdown",
            x=returns_df.index,
            y=returns_df["Drawdown"],
            line={"shape": "hv", "color": "salmon"},
            fill="tozeroy",
            fillcolor="rgba(255, 99, 71, 0.35)",
        ),
        row=3,
        col=1,
    )

    fig.update_yaxes(title_text="Drawdown", row=3, col=1)
    fig.update_xaxes(title_text="Datetime", row=3, col=1)
    distplot_bench = ff.create_distplot(
        [price_df["Returns"]],
        ["Benchmark Returns"],
        colors=["violet"],
        curve_type="kde",
        bin_size=3.5
        * price_df["Returns"].std()
        / (len(price_df["Returns"]) ** (1 / 3)),
    )
    fig.add_trace(distplot_bench["data"][0], row=4, col=1)
    distplot = ff.create_distplot(
        [returns_df["Returns"]],
        ["Strategy Returns"],
        colors=["salmon"],
        curve_type="kde",
        bin_size=3.5
        * returns_df["Returns"].std()
        / (len(returns_df["Returns"]) ** (1 / 3)),
    )
    fig.add_trace(distplot["data"][0], row=4, col=1)

    fig.update_xaxes(title_text="Returns", row=4, col=1)
    fig.update_yaxes(title_text="Density", row=4, col=1)
    fig.update_layout(
        xaxis_rangeslider_visible=False,
        showlegend=True,
        title_text="Historical price, strategy equity evolution/drawdown and returns distribution",
        height=1000,
    )
    fig.show()



import pandas as pd
import numpy as np
from ta.momentum import rsi, roc, ppo_hist
from ta.trend import ema_indicator


class TechnicalScreener:
    @staticmethod
    def __calculate_indicators(df: pd.DataFrame) -> pd.DataFrame:
        """Calculate all indicators needed to do the technical scoring.

        Args:
        ----
            df (pd.DataFrame): The dataframe to calculate the indicators on that needs to have the following columns: Open, High, Low, Close, Volume.

        Returns:
        ----
            pd.DataFrame: The dataframe with the calculated indicators.
        """
        assert df.shape[0] > 1, "Dataframe must have at least 2 rows"
        assert set({"Open", "High", "Low", "Close"}).issubset(
            df.columns
        ), "Dataframe must have columns Open, High, Low, Close"
        df_clone = df.copy()
        #  Long-term
        df_clone["EMA_200"] = ema_indicator(df_clone["Close"], 200, fillna=True)
        df_clone["EMA_200_CLOSE_PC"] = (df_clone["Close"] / df_clone["EMA_200"]) * 100
        df_clone["ROC_125"] = roc(df_clone["Close"], 125, fillna=True)
        #  Mid-term
        df_clone["EMA_50"] = ema_indicator(df_clone["Close"], 50, fillna=True)
        df_clone["EMA_50_CLOSE_PC"] = (df_clone["Close"] / df_clone["EMA_50"]) * 100
        df_clone["ROC_20"] = roc(df_clone["Close"], 20, fillna=True)
        # Short-term
        df_clone["PPO_HIST"] = ppo_hist(
            df_clone["Close"],
            window_slow=26,
            window_fast=12,
            window_sign=9,
            fillna=True,
        )
        #  Calculate PPO histogram slope
        df_clone["PPO_HIST_SLOPE"] = (
            df_clone["PPO_HIST"]
            .rolling(3)
            .apply(
                lambda rows: np.polyfit(
                    np.arange(1, len(rows) + 1, 1), np.array(rows.values), 1
                )[0]
            )
            .fillna(0)
        )
        df_clone["RSI"] = rsi(df_clone["Close"], window=14, fillna=True)
        return df_clone

    @staticmethod
    def __calculate_weights(dataframe: pd.DataFrame) -> pd.Series:
        """Calculate the indicators and the weights for each indicator.

        Args:
        ----
            dataframe (pd.DataFrame): The dataframe to calculate the indicators on that needs to have the following columns: Open, High, Low, Close, Volume.

        Returns:
        ----
            pd.Series: The calculated weighted score.
        """
        df = TechnicalScreener.__calculate_indicators(dataframe)
        #  Long-term
        df["EMA_200_CLOSE_PC_WEIGHTED"] = df["EMA_200_CLOSE_PC"] * 0.3
        df["ROC_125_WEIGHTED"] = df["ROC_125"] * 0.3
        #  Mid-term
        df["EMA_50_CLOSE_PC_WEIGHTED"] = df["EMA_50_CLOSE_PC"] * 0.15
        df["ROC_20_WEIGHTED"] = df["ROC_20"] * 0.15
        #  Short-term
        df["RSI_WEIGHTED"] = df["RSI"] * 0.05
        df["PPO_HIST_SLOPE_WEIGHTED"] = 0
        df.loc[df["PPO_HIST_SLOPE"] < -1, "PPO_HIST_SLOPE_WEIGHTED"] = 0
        df.loc[df["PPO_HIST_SLOPE"] >= -1, "PPO_HIST_SLOPE_WEIGHTED"] = (
            (df["PPO_HIST_SLOPE"] + 1) * 50 * 0.05
        )
        df.loc[df["PPO_HIST_SLOPE"] > 1, "PPO_HIST_SLOPE_WEIGHTED"] = 5

        return (
            df["EMA_200_CLOSE_PC_WEIGHTED"]
            + df["ROC_125_WEIGHTED"]
            + df["EMA_50_CLOSE_PC_WEIGHTED"]
            + df["ROC_20_WEIGHTED"]
            + df["RSI_WEIGHTED"]
            + df["PPO_HIST_SLOPE_WEIGHTED"]
        )

    @classmethod
    def score(cls, df: pd.DataFrame) -> pd.Series:
        return cls.__calculate_weights(dataframe=df)



from sklearn.decomposition import PCA
from typing import Iterable, Literal, Optional
import pandas as pd
import numpy as np
import numpy.typing as npt
import numpy.linalg as linalg
import matplotlib.pyplot as plt
from tqdm import tqdm
from abc import ABC, abstractmethod
from scipy.optimize import minimize
import plotly.graph_objects as go
import plotly.express as px


class ABCPortfolio(ABC):
    def __init__(self, returns: pd.DataFrame, trading_days: int = 365) -> None:
        """Construct a new 'ABCPortfolio' object. Generic for all portfolio optimization models : `MonteCarloPortfolio`, `ConvexPortfolio`.

        Args:
        -----
            returns (pd.DataFrame): The returns of the assets in the portfolio.
            trading_days (int, optional): The number of trading days in a year, 365 for cryptos, 252 for stocks. Defaults to 365.
        """
        self._returns = returns
        self._returns_mean = self._returns.mean()
        self._returns_cov = self._returns.cov().values
        self._trading_days = trading_days
        self._already_optimized = False

    @abstractmethod
    def fit(self):
        pass

    @abstractmethod
    def get_allocation(self):
        pass

    @staticmethod
    def _plot_allocation(
        weights: npt.NDArray | Iterable,
        assets: npt.NDArray | list[str] | tuple[str] | pd.Index,
    ) -> None:
        """Plot the allocation of the portfolio in  a pie chart.

        Args:
        -----
            weights (npt.NDArray | Iterable): The weights for each asset in the portfolio between 0 and 1.
            assets (npt.NDArray | list[str] | tuple[str]): The assets (their names) in the portfolio.
        """
        res = [(p, float(w)) for p, w in zip(assets, weights)]
        res.sort(key=lambda x: x[1], reverse=True)
        res = np.array(res)
        plt.title("Asset allocation")
        plt.pie(res[:, -1], labels=res[:, 0].tolist(), autopct="%1.1f%%")
        plt.show()
        # fig = go.Figure(
        #     go.Pie(
        #         labels=res[:, 0].tolist(),
        #         values=res[:, -1],
        #         hole=0.4,
        #         text=[f"{p}: {float(w):.2f} %" for p, w in res],
        #         textinfo="label+percent",
        #         hoverinfo="label+value",
        #         marker=dict(colors=px.colors.qualitative.Pastel1),
        #     )
        # )

        # fig.update_layout(
        #     title={
        #         "text": "Asset allocation",
        #         "x": 0.5,
        #         "y": 0.95,
        #         "xanchor": "center",
        #         "yanchor": "top",
        #     },
        #     showlegend=False,
        #     height=600,
        #     width=600,
        # )

        # fig.show()

    def _compute_metrics(
        self,
        weights: npt.NDArray,
    ) -> dict[str, float | int]:
        """Given an array of weights, compute the Sharpe ratio, the risk and the return of the portfolio. This method in used it the optimization functions of the portfolio optimization models.

        Args:
        -----
            weights (npt.NDArray): The weights of the assets in the portfolio.

        Returns:
        -----
            dict[str, float | int]: The Sharpe ratio, the risk and the return of the portfolio. The keys are : "sharpe", "risk" and "return".
        """
        ret = np.sum(self._returns_mean * weights * self._trading_days)
        vol = float(
            np.sqrt(weights.T @ self._returns_cov * self._trading_days @ weights)
        )
        sr = ret / vol

        co = pd.DataFrame(self._returns.values * weights).corr()
        np.fill_diagonal(co.values, 1)
        co.fillna(0, inplace=True)
        div = float(linalg.norm(co - np.eye(co.shape[1]), ord="fro"))

        return {"sharpe": sr, "risk": vol, "return": ret, "ptf_correlation": div}


class RiskParityPortfolio(ABCPortfolio):
    def __compute_objective_metrics(self, weights: npt.NDArray, args: list) -> float:
        """The objective function to optimize. It is used in the `fit` method.

        Args:
        -----
            weights (npt.NDArray): The weights of the assets in the portfolio.
            args (list): The arguments of the objective function.

        Returns:
        -----
            float: The value of the objective function.
        """

        # We convert the weights to a matrix
        weights_matrix = np.matrix(weights)
        assets_risk_budget = np.matrix(args[0])

        # We calculate the risk of the weights distribution
        portfolio_risk = self._compute_metrics(weights=weights)["risk"]

        # We calculate the contribution of each asset to the risk of the weights
        # distribution
        assets_risk_contribution = (
            np.multiply(
                weights_matrix.T,
                self._returns_cov * self._trading_days * weights_matrix.T,
            )
            / portfolio_risk
        )

        # We calculate the desired contribution of each asset to the risk of the weights distribution
        assets_risk_target = portfolio_risk * assets_risk_budget

        # Error between the desired contribution and the calculated contribution of each asset
        return np.sum(np.square(assets_risk_contribution - assets_risk_target.T))  # MSE

    def fit(
        self,
    ) -> None:
        """Optimize the portfolio for an equity risk parity among assets. It uses an optimizer from `scipy.optimize` module. This method returns nothing."""

        cons = (
            {
                "type": "eq",
                "fun": lambda weights: np.sum(weights) - 1,
            },  # return 0 if sum of the weights is 1
            {"type": "ineq", "fun": lambda x: x},  # Long only
        )

        bounds = tuple([(0.0, 1.0) for _ in range(len(self._returns.columns))])
        init_guess = [
            1 / len(self._returns.columns) for _ in range(len(self._returns.columns))
        ]

        assets_risk_budget = [
            1 / len(self._returns.columns) for _ in range(len(self._returns.columns))
        ]

        opt_results = minimize(
            fun=self.__compute_objective_metrics,
            x0=init_guess,
            args=[assets_risk_budget],
            method="SLSQP",
            bounds=bounds,
            constraints=cons,
            tol=1e-10,
        )

        self._already_optimized = True
        self.__optimized_weights = opt_results.x
        self.__optimized_metrics = self._compute_metrics(self.__optimized_weights)

    def get_allocation(self) -> pd.DataFrame:
        """Plot the allocation of the portfolio and return a DataFrame with the allocation of each asset. This function can't be called before the fit method.

        Returns:
        --------
            pd.DataFrame: The allocation of each asset in the portfolio, columns are the assets and rows are the weights.
        """
        assert (
            self._already_optimized
        ), "You must fit the model before getting the allocation."

        print(f"{'  Results  ':-^40}")

        print(
            f"- Annualized Sharpe ratio: {self.__optimized_metrics.get('sharpe',0.0):.2f}\n- Annualized risk (volatility): {100*self.__optimized_metrics.get('risk',1.0):.2f} %\n- Annualized expected return: {100*self.__optimized_metrics.get('return',0.0):.2f} %"
        )
        ConvexPortfolio._plot_allocation(
            self.__optimized_weights, self._returns.columns
        )
        return pd.DataFrame(
            {
                p: [w]
                for p, w in zip(
                    self._returns.columns,
                    self.__optimized_weights,
                )
            }
        )


class ConvexPortfolio(ABCPortfolio):
    def __compute_objective_metrics(
        self, weights: npt.NDArray, args: list[str]
    ) -> float:
        """The objective function to optimize. It is used in the `fit` method.

        Args:
        -----
            weights (npt.NDArray): The weights of the assets in the portfolio.
            args (list[str]): The arguments of the objective function.

        Returns:
        -----
            float: The value of the objective function.
        """
        metric: Literal["sharpe", "risk", "return"] = args[0]  # type: ignore
        way: Literal["min", "max"] = args[1]  # type: ignore
        metrics = self._compute_metrics(weights)
        return -metrics.get(metric, 0.0) if way == "max" else metrics.get(metric, 0.0)

    def fit(
        self,
        metric: Literal["sharpe", "risk", "return"] = "sharpe",
        way: Literal["min", "max"] = "max",
        max_asset_weight: float = 0.3,
        min_asset_weight: float = 0.0,
    ) -> None:
        """Optimize the portfolio using the Sharpe ratio, risk or return as a metric and a maximization or a minimization this metric. It uses an optimizer from `scipy.optimize` module. This method returns nothing.

        Args:
        -----
            metric (Literal[&quot;sharpe&quot;, &quot;risk&quot;, &quot;return&quot;], optional): The metric to optimize. Defaults to "sharpe".
            way (Literal[&quot;min&quot;, &quot;max&quot;], optional): The type of wanted optimization optimization. Defaults to "max".
            max_asset_weight (float, optional): The maximal weight of an asset in the portfolio. Defaults to 0.3.
            min_asset_weight (float, optional): The minimal weight of an asset in the portfolio. Defaults to 0.0.
        """
        assert way in [
            "min",
            "max",
        ], "Invalid way of metric evaluation, must be a string equals to min or max."
        assert metric in [
            "sharpe",
            "risk",
            "return",
        ], "Invalid metric, must be a string equals to sharpe, risk or return."
        assert max_asset_weight <= 1.0, "Max asset weight must be less or equal to 1.0"
        assert (
            min_asset_weight >= 0.0
        ), "Min asset weight must be greater or equal to 0.0"
        assert (
            max_asset_weight >= min_asset_weight
        ), "Max asset weight must be greater or equal to min asset weight."

        cons = (
            {
                "type": "eq",
                "fun": lambda weights: np.sum(weights) - 1,
            },  # return 0 if sum of the weights is 1
            {"type": "ineq", "fun": lambda x: x},  # Long only
        )

        bounds = tuple(
            [
                (min_asset_weight, max_asset_weight)
                for _ in range(len(self._returns.columns))
            ]
        )
        init_guess = [
            1 / len(self._returns.columns) for _ in range(len(self._returns.columns))
        ]

        opt_results = minimize(
            fun=self.__compute_objective_metrics,
            x0=init_guess,
            args=[metric, way],
            method="SLSQP",
            bounds=bounds,
            constraints=cons,
            tol=1e-10,
        )

        self._already_optimized = True
        self.__optimized_weights = opt_results.x
        self.__optimized_metrics = self._compute_metrics(self.__optimized_weights)

    def get_allocation(self) -> pd.DataFrame:
        """Plot the allocation of the portfolio and return a DataFrame with the allocation of each asset. This function can't be called before the fit method.

        Returns:
        --------
            pd.DataFrame: The allocation of each asset in the portfolio, columns are the assets and rows are the weights.
        """
        assert (
            self._already_optimized
        ), "You must fit the model before getting the allocation."

        print(f"{'  Results  ':-^40}")

        print(
            f"- Annualized Sharpe ratio: {self.__optimized_metrics.get('sharpe',0.0):.2f}\n- Annualized risk (volatility): {100*self.__optimized_metrics.get('risk',1.0):.2f} %\n- Annualized expected return: {100*self.__optimized_metrics.get('return',0.0):.2f} %"
        )
        ConvexPortfolio._plot_allocation(
            self.__optimized_weights, self._returns.columns
        )
        return pd.DataFrame(
            {
                p: [w]
                for p, w in zip(
                    self._returns.columns,
                    self.__optimized_weights,
                )
            }
        )


class MonteCarloPortfolio(ABCPortfolio):
    def fit(self, n_portfolios: int = 20000, plot: bool = True):
        """This method run a Monte Carlo simulation to allocate and create many portfolios in order to then find the most efficient. This method returns nothing.

        Args:
        -----
            n_portfolios (int, optional): The number of portefolio to create. Defaults to 20000.
            trading_days (int, optional): The number of trading days in a year, 365 for cryptos, 252 for stocks. Defaults to 365.
            plot (bool, optional): Whether or not to plot the portfolios simulated on a chart, x-axis = risk, y-axis = return. Defaults to True.
        """
        self.__all_weights = np.zeros((n_portfolios, len(self._returns.columns)))
        self.__ret_arr = np.zeros(n_portfolios)
        self.__vol_arr = np.zeros(n_portfolios)
        self.__sharpe_arr = np.zeros(n_portfolios)

        for x in tqdm(range(n_portfolios)):
            weights = np.array(
                np.random.rand(len(self._returns.columns)), dtype=np.float64
            )
            weights = weights / np.sum(weights)

            # Save weights
            self.__all_weights[x, :] = weights

            metrics = self._compute_metrics(weights)

            # Expected return
            self.__ret_arr[x] = metrics[
                "return"
            ]  # np.sum((rets * weights * self._trading_days))

            # Expected volatility
            self.__vol_arr[x] = metrics[
                "risk"
            ]  # np.sqrt( weights.T @ cov_matrix * self._trading_days @ weights)

            # Sharpe Ratio
            self.__sharpe_arr[x] = metrics[
                "sharpe"
            ]  # self.__ret_arr[x] / self.__vol_arr[x]
        self._already_optimized = True

        if plot:
            self.plot_portfolios(n_portfolios)

    def plot_portfolios(self, n_portfolios: Optional[int] = None) -> None:
        """Plot the portfolios simulated on a chart, x-axis = risk, y-axis = return.

        Args:
        -----
            n_portfolios (Optional[int], optional): The number of portfolio. Defaults to None.

        """
        assert (
            self._already_optimized
        ), "You must fit the model before getting the allocation."

        fig = go.Figure(
            go.Scatter(
                x=self.__vol_arr,
                y=self.__ret_arr,
                mode="markers",
                marker=dict(
                    color=self.__sharpe_arr,
                    colorbar=dict(title="Sharpe Ratio"),
                    colorscale="viridis",
                ),
                text=[
                    f"Volatility: {vol:.2f}<br>Return: {ret:.2f}<br>Sharpe Ratio: {sharpe:.2f}"
                    for vol, ret, sharpe in zip(
                        self.__vol_arr, self.__ret_arr, self.__sharpe_arr
                    )
                ],
            )
        )

        fig.update_layout(
            title={
                "text": f"Monte Carlo Portfolio Optimization, {n_portfolios} simulations"
                if n_portfolios
                else "Monte Carlo Portfolio Optimization",
                "x": 0.5,
                "y": 0.95,
                "xanchor": "center",
                "yanchor": "top",
            },
            xaxis_title="Volatility",
            yaxis_title="Return",
            hovermode="closest",
            showlegend=False,
            height=650,
        )

        fig.add_trace(
            go.Scatter(
                x=[self.__vol_arr[self.__sharpe_arr.argmax()]],
                y=[self.__ret_arr[self.__sharpe_arr.argmax()]],
                mode="markers",
                marker=dict(color="red", size=10),
                name="Maximum Sharpe Ratio",
            )
        )

        fig.show()

    def get_allocation(
        self,
        metric: Literal["sharpe", "risk", "return"] = "sharpe",
        way: Literal["min", "max"] = "max",
    ) -> pd.DataFrame:
        """Plot the allocation of the portfolio and return a DataFrame with the allocation of each asset. This function can't be called before the fit method.

        Args:
        -----
            metric (Literal[&quot;sharpe&quot;, &quot;risk&quot;, &quot;return&quot;], optional): The metric to optimize. Defaults to "sharpe".
            way (Literal[&quot;min&quot;, &quot;max&quot;], optional): The type of wanted optimization optimization. Defaults to "max".

        Raises:
        -------
            Exception: Invalid metric chosen.

        Returns:
        --------
            pd.DataFrame: The allocation of each asset in the portfolio, columns are the assets and rows are the weights.
        """
        assert (
            self._already_optimized
        ), "You must fit the model before getting the allocation."
        assert way in [
            "min",
            "max",
        ], "Invalid way of metric evaluation, must be a string equals to min or max."

        match metric:
            case "sharpe":
                ind = (
                    self.__sharpe_arr.argmin()
                    if way == "min"
                    else self.__sharpe_arr.argmax()
                )
            case "risk":
                ind = (
                    self.__vol_arr.argmin() if way == "min" else self.__vol_arr.argmax()
                )
            case "return":
                ind = (
                    self.__vol_arr.argmin() if way == "min" else self.__vol_arr.argmax()
                )
            case _:
                raise Exception("Invalid metric.")

        print(f"{'  Results  ':-^40}")
        print(
            f"- Annualized Sharpe ratio: {self.__sharpe_arr[ind]:.2f}\n- Annualized risk (volatility): {100*self.__vol_arr[ind]:.2f} %\n- Annualized expected return: {100*self.__ret_arr[ind]:.2f} %"
        )
        MonteCarloPortfolio._plot_allocation(
            self.__all_weights[ind, :], self._returns.columns
        )
        return pd.DataFrame(
            {
                p: [w]
                for p, w in zip(
                    self._returns.columns,
                    self.__all_weights[ind, :],
                )
            }
        )


def reduce_dimentionality(
    full_dataframe: pd.DataFrame,
    mode: Literal["pca", "tsne"] = "pca",
    target_explained_variance: float = 0.95,
    result_as_df: bool = True,
) -> npt.NDArray[np.float64] | pd.DataFrame:
    """Reduce the dimension of a dataset given a target explained variance.

    Args:
    -----
        full_dataframe (pd.DataFrame): The dataframe to reduce the dimension.

        mode (Literal[&quot;pca&quot;, &quot;tsne&quot;], optional): The dimensionality reduction algorithm. Defaults to "pca".

        target_explained_variance (float, optional): The minimum threshold for explained variance. Defaults to 0.95.

        result_as_df (bool, optional): Whether or not return the result as a dataframe or as a numpy array. Defaults to True.

    Returns:
    -------
        npt.NDArray[np.float64] | pd.DataFrame: The reduced dataset.
    """
    assert (
        isinstance(full_dataframe, pd.DataFrame)
        and full_dataframe.shape[0] != 0
        and full_dataframe.shape[1] > 1
    ), "full_dataframe must be a pandas DataFrame containing at least 2 columns and 1 row"
    assert (
        target_explained_variance > 0 and target_explained_variance < 1
    ), "target_explained_variance must be a float between 0 and 1"
    X = full_dataframe.to_numpy()
    match mode:
        case "pca":
            calibration_pca = PCA()
            calibration_pca.fit(X)

            d = (
                np.argmax(
                    np.cumsum(calibration_pca.explained_variance_ratio_)
                    >= target_explained_variance
                )
                + 1
            )
            pca = PCA(n_components=d)
            pca.fit(X)

            if result_as_df:
                return pd.DataFrame(
                    pca.transform(X), columns=[f"PC_{i}" for i in range(1, d + 1)]
                )
            return pca.transform(X)
        case "tsne":
            raise NotImplementedError("TSNE is not implemented yet")
        case _:
            raise ValueError("mode must be either pca or tsne")



import numpy as np
import numpy.typing as npt
import pandas as pd
import scipy.stats as stat
from typing import Literal, Union


def sharpe_ratio(
    returns: pd.Series,
    N: Union[int, float] = 365,
    risk_free_rate: float = 0.03,
) -> float:
    """The economist William F. Sharpe proposed the Sharpe ratio in 1966 as an extension of his work on the Capital Asset Pricing Model (CAPM). It is defined as the difference between the returns of the investment and the risk-free return, divided by the standard deviation of the investment.

    Args:
    -----
        returns (pd.Series): The strategy or portfolio not cumulative returns.

        N (Union[int, float], optional): The number of periods in a year. Defaults to 365.

        risk_free_rate (float, optional): The risk free rate usually 10-year bond, buy-and-hold or 0. Defaults to 0.0.

    Returns:
    -----
        float: The annualized sharpe ratio.
    """
    return (returns.mean() * N - risk_free_rate) / (returns.std() * (N**0.5))


def treynor_ratio(
    returns: pd.Series,
    benchmark_returns: pd.Series,
    N: Union[int, float] = 365,
    risk_free_rate: float = 0.03,
) -> float:
    """The Treynor ratio is a risk-adjusted measure of return based on systematic risk. It is similar to the Sharpe ratio, except it uses beta as the measurement of volatility instead of standard deviation.

    Args:
    -----
        returns (pd.Series): The strategy or portfolio not cumulative returns.

        benchmark_returns (pd.Series): The strategy or portfolio benchmark not cumulative returns.

        N (Union[int, float], optional): The number of periods in a year. Defaults to 365.

        risk_free_rate (float, optional): The risk free rate usually 10-year bond, buy-and-hold or 0. Defaults to 0.0.

    Returns:
    -----
        float: The annualized treynor ratio.
    """
    beta = np.cov(returns, benchmark_returns)[0, 1] / np.var(returns)
    return (returns.mean() * N - risk_free_rate) / (beta * (N**0.5))


def sortino_ratio(
    returns: pd.Series,
    N: Union[int, float] = 365,
    risk_free_rate: float = 0.03,
) -> float:
    """The Sortino ratio is very similar to the Sharpe ratio, the only difference being that where the Sharpe ratio uses all the observations for calculating the standard deviation the Sortino ratio only considers the harmful variance.

    Args:
    -----
        returns (pd.Series): The strategy or portfolio not cumulative returns.

        N (Union[int, float], optional): The number of periods in a year. Defaults to 365.

        risk_free_rate (float, optional): The risk free rate usually 10-year bond, buy-and-hold or 0. Defaults to 0.0.

    Returns:
    -----
        float: The annualized sortino ratio.
    """
    return (returns.mean() * N - risk_free_rate) / (downside_risk(returns) * (N**0.5))


def calmar_ratio(
    returns: pd.Series,
    N: Union[int, float] = 365,
) -> float:
    """The final risk/reward ratio we will consider is the Calmar ratio. This is similar to the other ratios, with the key difference being that the Calmar ratio uses max drawdown in the denominator as opposed to standard deviation.

    Args:
        returns (pd.Series): The strategy or portfolio not cumulative returns.

        N (Union[int, float], optional): The number of periods in a year. Defaults to 365.

    Returns:
        float: The annualized calmar ratio.
    """
    return (returns.mean() * N) / abs(max_drawdown(returns))


def downside_risk(returns: pd.Series) -> float:
    """Downside risk or Semi-Deviation is a method of measuring the fluctuations below the mean, unlike variance or standard deviation it only looks at the negative price fluctuations and it's used to evaluate the downside risk (The risk of loss in an investment) of an investment.

    Args:
    -----
        returns (pd.Series): The strategy or portfolio not cumulative returns.

    Returns:
    ------
        float: The semi-deviation or downside risk of returns.
    """
    return returns.loc[returns < 0].std()


def drawdown(returns: pd.Series) -> pd.Series:
    """Computes the drawdown series of a given returns (not cumulative) time series.

    Args:
    ----
        returns (pd.Series): The strategy or portfolio not cumulative returns.

    Returns:
    ----
        pd.Series: The drawdown series.
    """
    cumulative_returns = (returns + 1).cumprod()
    running_max = cumulative_returns.cummax()
    return (cumulative_returns - running_max) / running_max


def max_drawdown(
    returns: pd.Series,
) -> float:
    """Max drawdown quantifies the steepest decline from peak to trough observed for an investment.

    Args:
    ----
        returns (pd.Series): The strategy or portfolio not cumulative returns.

    Returns:
    ----
        float: The max drawdown.
    """
    return drawdown(returns).min()


def information_ratio(
    portfolio_returns: pd.Series, benchmark_returns: pd.Series
) -> float:
    """The information ratio (IR) is a measurement of portfolio returns beyond the returns of a benchmark, usually an index, compared to the volatility of those returns. The information ratio (IR) measures a portfolio manager's ability to generate excess returns relative to a benchmark but also attempts to identify the consistency of the investor.

    Args:
    -----
        portfolio_returns (pd.Series): The strategy or portfolio not cumulative returns.

        benchmark_returns (pd.Series): The strategy or portfolio benchmark not cumulative returns.

    Returns:
    -----
        float: The annualized information ratio.
    """
    return (portfolio_returns - benchmark_returns).mean() / tracking_error(
        portfolio_returns, benchmark_returns
    )


def tracking_error(portfolio_returns: pd.Series, benchmark_returns: pd.Series) -> float:
    """Tracking error is the divergence between the price behavior of a position or a portfolio and the price behavior of a benchmark.

    Args:
    -----
        portfolio_returns (pd.Series): The strategy or portfolio not cumulative returns.

        benchmark_returns (pd.Series): The strategy or portfolio benchmark not cumulative returns.

    Returns:
    -----
        float: The annualized tracking error.
    """
    return (portfolio_returns - benchmark_returns).std()


def kelly_criterion(returns: pd.Series) -> float:
    p = (returns > 0).mean()
    q = 1 - p
    win = returns[returns > 0].mean()
    loss = returns[returns < 0].mean()
    r = win / abs(loss)
    return float((p * r - q) / r)


def value_at_risk(
    returns: pd.Series,
    level: int = 5,
    method: Literal["historic", "gaussian", "cornish_fischer"] = "historic",
) -> float:
    """Returns the VaR of a Series or DataFrame using the specified method.

    Args:
    -----
        returns (pd.Series): The strategy or portfolio not cumulative returns.

        level (int, optional): Percentile to compute, which must be between 0 and 100 inclusive. Defaults to 5.

        method (Literal[&quot;historic&quot;, &quot;gaussian&quot;, &quot;cornish_fischer&quot;], optional): The method for VaR calculation : historic use the returns provided, gaussian will approximate the returns by a a gaussian parametric distribution and will correct the gaussian VaR using skewness and kurtosis. Defaults to "historic".

    Returns:
    -----
        float: The historical VaR.
    """
    if method == "historic":
        return float(np.percentile(returns.to_numpy(), level))
    elif method == "gaussian":
        return float((returns.mean() + stat.norm.ppf(level / 100) * returns.std()))
    elif method == "cornish_fischer":
        z = stat.norm.ppf(level / 100)
        s = stat.skew(returns.values)
        k = stat.kurtosis(returns.values)

        return float(
            (
                returns.mean()
                + (
                    z
                    + (z**2 - 1) * s / 6
                    + (z**3 - 3 * z) * (k - 3) / 24
                    - (2 * z**3 - 5 * z) * (s**2) / 36
                )
                * returns.std()
            )
        )
    else:
        raise ValueError(
            "VaR calculation method must be historic, gaussian or cornish_fischer"
        )


def conditional_value_at_risk(
    returns: pd.Series,
    level: int = 5,
    method: Literal["historic", "gaussian", "cornish_fischer"] = "historic",
) -> float:
    """Returns the CVaR (conditional value-at-risk) also called expected shortfall of a Series or DataFrame

    Args:
    ----
        returns (pd.Series): The strategy or portfolio not cumulative returns.

        level (int, optional): Percentile to compute, which must be between 0 and 100 inclusive. Defaults to 5.

        method (Literal[&quot;historic&quot;, &quot;gaussian&quot;, &quot;cornish_fischer&quot;], optional): The method for VaR calculation : historic use the returns provided, gaussian will approximate the returns by a a gaussian parametric distribution and will correct the gaussian VaR using skewness and kurtosis. Defaults to "historic".

    Returns:
    ----
        float: The CVaR.
    """
    if method == "historic":
        historic_CVaR = np.mean(
            returns[returns < value_at_risk(returns, level, "historic")].to_numpy()
        )
        return (
            float(historic_CVaR)
            if not np.isnan(historic_CVaR)
            else value_at_risk(returns, level)
        )
    else:
        raise NotImplementedError("Only historic CVaR is currently implemented")



import pandas as pd
import numpy as np
import numpy.typing as npt
from typing import Literal, Callable, Optional
from scipy import stats, signal
from scipy.signal import hilbert
from ta.trend import ema_indicator, sma_indicator
from ta.momentum import ppo_hist
from ta.volatility import average_true_range


def dynamic_time_wrapping_indicator(
    timeseries_1: pd.Series, timeseries_2: pd.Series, window: int = 20
) -> pd.Series:
    """Calculate the dynamic time warping cost over all dates possible. Inspired from this article : https://medium.datadriveninvestor.com/one-cool-way-to-spot-correlations-in-finance-dynamic-time-warping-763a6e7f0ee3

    Args:
    -----
        timeseries_1 (pd.Series): The first time series.

        timeseries_2 (pd.Series): The second time series.

        window (int, optional): How long the dynamic time-warping comparison is. Defaults to 20.

    Returns:
    -----
        pd.Series: The dynamic time warping cost indicator over all dates possible.
    """

    def get_dynamic_time_wrapping_score(
        ts1: npt.NDArray[np.float64], ts2: npt.NDArray[np.float64]
    ) -> float:
        """Calculate the dynamic time warping cost between two time series

        Args:
        -----
            ts1 (npt.NDArray[np.float64]): The first time series.

            ts2 (npt.NDArray[np.float64]): The second time series.

        Returns:
        -----
            float: The dynamic time warping cost between the two time series.
        """
        C = np.full(
            shape=(ts1.shape[0] + 1, ts2.shape[0] + 1),
            fill_value=np.inf,
        )  # Initialise a full cost matrix, filled with np.inf. This is so we can start the algorithm and not get stuck on the boundary

        C[
            0, 0
        ] = 0  # Place the corner to zero, so that we don't have the minimum of 3 infs

        for i in range(1, ts1.shape[0] + 1):
            for j in range(1, ts2.shape[0] + 1):
                # Euclidian distance between the two points
                dist = abs(ts1[i - 1] - ts2[j - 1])

                # Find the cheapest cost of all three neighbours
                prev_min = min(C[i - 1, j], C[i, j - 1], C[i - 1, j - 1])

                # Populate the entry in the cost matrix
                C[i, j] = dist + prev_min

        return C[-1, -1]

    z_scale: Callable[[pd.Series], npt.NDArray[np.float64]] = lambda ts: np.array(
        ((ts - ts.mean()) / ts.std()).values, dtype=np.float64
    )  # Lambda function to Z-scale the time series

    scores = (
        pd.Series(data=np.full(timeseries_1.shape, 1), index=timeseries_1.index)
        .rolling(window)
        .apply(
            lambda rows: get_dynamic_time_wrapping_score(
                ts1=z_scale(timeseries_1.loc[rows.index[0] : rows.index[-1]]),
                ts2=z_scale(timeseries_2.loc[rows.index[0] : rows.index[-1]]),
            )
        )
    )

    return scores.fillna(scores.min())


def vertical_horizontal_filter_indicator(
    close: pd.Series, window: int = 60
) -> pd.Series:
    """THe Vertical Horizontal Filter (VHF) is a technical analysis indicator used to determine whether prices are in a trending or non-trending phase. It was introduced by Adam White in an article in the August, 1991 issue of Futures Magazine. The VHF uses the highest high and lowest low of the last n periods to determine whether the market is in a trending phase or a congestion phase.

    Args:
    ----
        close (pd.Series): The close series from the OHLCV Data.

        window (int, optional): The indicator window for the sum and min/max. Defaults to 60.

    Returns:
    ----
        pd.Series: The Vertical Horizontal Filter indicator.
    """
    return (
        close.rolling(window).max() - close.rolling(window).min()
    ) / close.diff().fillna(0).apply(lambda x: abs(x)).rolling(window).sum()


def extreme_euphoria_pattern(
    open: pd.Series,
    close: pd.Series,
    euphoria_side: Literal["bull", "bear"] = "bull",
    n_candles: int = 3,
):
    assert (
        n_candles > 0 and n_candles < 10
    ), "n_candles must be greater than 0 and less than 10"
    open_close_return = close - open  # (close/open) - 1

    def euphoria(close_open: pd.Series) -> int:
        close_open_abs = close_open.abs()
        if euphoria_side == "bear":
            if np.all(np.sign(close_open.to_numpy()) == 1) and all(
                close_open_abs.iloc[i] <= close_open_abs.iloc[i + 1]
                for i in range(len(close_open_abs) - 1)
            ):
                return 0
            return 1
        elif euphoria_side == "bull":
            if np.all(np.sign(close_open.to_numpy()) == -1) and all(
                close_open_abs.iloc[i] <= close_open_abs.iloc[i + 1]
                for i in range(len(close_open_abs) - 1)
            ):
                return 1
            return 0
        else:
            raise ValueError("euphoria_side parameter must be either bull or bear")

    return open_close_return.rolling(n_candles).apply(euphoria)


def hilbert_transform_dominant_cycle_period_indicator(
    close: pd.Series, window: int
) -> pd.Series:
    analytic_signal = pd.Series(np.imag(hilbert(close)), index=close.index)
    instantaneous_phase = pd.Series(
        np.unwrap(np.angle(analytic_signal)), index=close.index
    )

    # Calculate the first differences of the instantaneous phase
    phase_diff = instantaneous_phase.diff().fillna(0)

    def htdcp_calculation(x: float) -> float:
        numerator = np.sum(np.arange(1, window + 1) * x)
        denominator = np.sum(x)
        return numerator / denominator if denominator != 0 else 0

    return phase_diff.rolling(window=window).apply(htdcp_calculation)


def chande_momentum_oscillator(
    high: pd.Series, low: pd.Series, window: int
) -> pd.Series:
    """The Chande oscillator is similar to other momentum indicators such as Wilder's relative strength index (RSI) and the stochastic oscillator. It measures momentum on both up and down days and does not smooth results, triggering more frequent oversold and overbought penetrations. The indicator oscillates between +100 and -100

    Args:
        high (pd.Series): The high series from the OHLCV Data.
        low (pd.Series): The low series from the OHLCV Data.
        window (int): The indicator window for the sum. Default to 20.

    Returns:
        pd.Series: The Chande momentum oscillator.
    """
    high_sum = high.rolling(window).sum()
    low_sum = low.rolling(window).sum()
    return 100 * (high_sum - low_sum) / (high_sum + low_sum)


def ulcer_index_indicator(close: pd.Series, high: pd.Series, window: int) -> pd.Series:
    """The Ulcer Index is a volatility indicator that measures downside risk. It was developed by Peter Martin and is based on the idea that investors are more concerned with the downside risk than the upside potential. The Ulcer Index is calculated as follows:
    UI = sqrt(sum[(P(i)/Max(P) — 1)², i=1 to N] / N) * 100
    """
    rolling_max_high = high.rolling(window).max()
    return (
        ((100 * (close - rolling_max_high) / rolling_max_high) ** 2) / window
    ) ** 0.5


def correlation_adjusted_reversal_indicator(
    close: pd.Series, window: int = 10
) -> pd.Series:
    """The idea for the correlation-adjusted reversal indicator is to detect average extremes where the correlation between returns and prices is high enough to justify a possible market inflection.

    Args:
        close (pd.Series): The close series from the OHLCV Data.
        window (int, optional): The indicator window for the correlation calculation. Defaults to 10.

    Returns:
        pd.Series: The correlation adjusted reversal indicator.
    """
    shifted_diff = pd.DataFrame()
    for i in range(1, window + 1):
        shifted_diff = pd.concat(
            [
                shifted_diff,
                pd.Series(close - close.shift(i), name=f"Shift_diff_{i}").fillna(0),
            ],
            axis=1,
        )

    return close.rolling(10).corr(shifted_diff.mean(axis=1)).fillna(0)


def garman_klass_volatility_indicator(
    open: pd.Series, close: pd.Series, high: pd.Series, low: pd.Series, window: int = 20
) -> pd.Series:
    """Calculate the Garman Klass volatility indicator.

    Args:
        open (pd.Series): The open series from the OHLCV Data.
        close (pd.Series): The close series from the OHLCV Data.
        high (pd.Series): The high series from the OHLCV Data.
        low (pd.Series): The low series from the OHLCV Data.
        window (int): The indicator window for the moving averages. Default to 20.

    Returns:
        pd.Series: The Garman Klass volatility indicator.
    """
    return ((np.log(high) - np.log(low)) ** 2) / 2 - (2 * np.log(2) - 1) * (
        (np.log(close) - np.log(open)) ** 2
    )


def amihud_illiquidity_indicator(close: pd.Series, volume: pd.Series) -> pd.Series:
    """Calculate the Amihud illiquidity indicator.

    Args:
        close (pd.Series): The close series from the OHLCV Data.
        volume (pd.Series): The volume series from the OHLCV Data.

    Returns:
        pd.Series: The Amihud illiquidity indicator.
    """
    return close.pct_change().abs() / (close * volume)


def equilibrium_indicator(close: pd.Series, window: int = 5) -> pd.Series:
    """Calculate equilibrum indicator aimed to detect mean reversion.

    Args:
        close (pd.Series): The close series from the OHLCV Data.
        window (int): The indicator window for the moving averages. Default to 5.

    Returns:
        pd.Series: The equilibrum indicator.
    """

    return ema_indicator(sma_indicator(close, window).fillna(method="backfill") - close)


def gap_indicator(open: pd.Series, close: pd.Series) -> pd.Series:
    """Calculate the % gap between the current open and the previous close.

    Args:
        open (pd.Series): The open series from the OHLCV Data.
        close (pd.Series): The close series from the OHLCV Data.

    Returns:
        pd.Series: The gap indicator.
    """

    def gap(opens: pd.Series) -> float:
        assert len(opens) == 2, "Error, open_rows must be a series of 2 values."
        prior_close = close.loc[opens.index[0]]
        current_open = opens[-1]
        return 100 * (current_open - prior_close) / prior_close

    return open.rolling(2).apply(gap)


def choppiness_index_indicator(
    high: pd.Series, low: pd.Series, close: pd.Series, window: int = 20
) -> pd.Series:
    """Compute the choppiness index indicator which indicates whether the market is trending, reversing or ranging. Boundaries are 38.2, 61.8.

    Args:
        high (pd.Series): The high series from the OHLCV Data.
        low (pd.Series): The low series from the OHLCV Data.
        close (pd.Series): The close series from the OHLCV Data.
        window (int, optional): The window to calculate the CHOP on a period. Defaults to 20.

    Returns:
        pd.Series: The choppiness index
    """
    return (
        100
        * np.log(
            average_true_range(high, low, close, window + 5).rolling(window).sum()
            / (high.rolling(window).max() - low.rolling(window).min())
        )
        / np.log(window)
    ).fillna(0)


def support_resistance_breakout_indicator(
    support_resistance_levels: list[float], open: pd.Series, low: pd.Series
) -> pd.Series:
    """Generate support resistance breakout rolling indicator.

    Args:
        support_resistance_levels (list[float]): The support and resistance levels.
        open (pd.Series[float | int]): The open series from the OHLCV Data.
        low (pd.Series[float | int]): The low series from the OHLCV Data.

    Returns:
        pd.Series: The S/R breakout rolling indicator.
    """

    def has_breakout(open_rows: pd.Series) -> int:
        """Base on the candle stick, detect a breakout or not.

        Args:
            open_rows (pd.Series): A series of 2 rolling open measures.

        Returns:
            int: 1 = breakout, 0 no breakout.
        """
        assert len(open_rows) == 2, "Error, open_rows must be a series of 2 values."
        last_index = open_rows.index[-1]
        for level in support_resistance_levels:
            if (
                (open_rows[0] < level)
                and (open_rows[-1] > level)
                and (low.loc[last_index] > level)
            ):
                return int(1)
        return int(0)

    return open.rolling(2).apply(has_breakout)


def zlema(
    close: pd.Series,
    window: int = 26,
) -> pd.Series:
    """ZLEMA is an abbreviation of Zero Lag Exponential Moving Average. It was developed by John Ehlers and Rick Way.
    ZLEMA is a kind of Exponential moving average but its main idea is to eliminate the lag arising from the very nature of the moving averages
    and other trend following indicators. As it follows price closer, it also provides better price averaging and responds better to price swings.

    Args:
        close (pd.Series): The close series from the OHLCV Data.
        window (int, optional): The window to calculate the zlema on a period. Defaults to 26.

    Returns:
        pd.Series: The ZLEMA indicator
    """
    return (close + (close.diff((window - 1) // 2))).ewm(span=window).mean()


def fibonacci_moving_average(
    close: pd.Series, fibonacci_n_terms: int = 17
) -> pd.Series:
    terms = tuple(
        [
            2,
            3,
            5,
            8,
            13,
            21,
            34,
            55,
            89,
            144,
            233,
            377,
            610,
            987,
            1597,
            2584,
            4181,
            6765,
            10946,
        ]
    )
    if fibonacci_n_terms > len(terms):
        raise ValueError(
            "Error, provide less fibonacci terms, it must be lower that 17."
        )

    temp_df = pd.DataFrame({"Close": close.values})
    for i in range(fibonacci_n_terms):
        temp_df[f"EMA{terms[i]}"] = temp_df["Close"].ewm(terms[i]).mean()
    temp_df.drop(columns=["Close"], inplace=True)
    return temp_df.apply(lambda row: row.mean(), axis=1)


def fibonacci_distance_indicator(
    close: pd.Series,
    fibonacci_ma: pd.Series,
    normalization: bool = True,
    normalization_window: int = 20,
    ma_smoothing: bool = True,
    ma_smoothing_window: int = 10,
) -> pd.Series:
    """_summary_

    Args:
        close (pd.Series): The market, i.e. often the closing price.
        fibonacci_ma (pd.Series): The pre-computed fibonacci moving average.
        normalization (bool, optional): Whether to normalize the distance indicator with min-max scaling. Defaults to True.
        normalization_window (int, optional): The min-max scaling window. Defaults to 20.
        ma_smoothing (bool, optional):  Whether to smooth or not the distance indicator. True includes the normalization step. Defaults to True.
        ma_smoothing_window (int, optional): The smoothing window. Defaults to 10.

    Returns:
        pd.Series: The fibonacci distance indicator.
    """
    dist = close - fibonacci_ma
    if normalization is True or ma_smoothing is True:
        dist_normalized = (dist - dist.rolling(normalization_window).min()) / (
            dist.rolling(normalization_window).max()
            - dist.rolling(normalization_window).min()
        )
        if ma_smoothing is True:
            return dist_normalized.rolling(ma_smoothing_window).mean()
        else:
            return dist_normalized
    return dist


def fischer_transformation(close: pd.Series, window: int = 5) -> pd.Series:
    """Compute the fischer transformation on a series.

    Args:
        close (pd.Series): The series to transform.
        window (int): The window to compute the fischer transformation on. Default to 5.

    Returns:
        pd.Series: The transformed series.
    """
    close_normalized: pd.Series = (close - close.rolling(window=window).min()) / (
        close.rolling(window=window).max() - close.rolling(window=window).min()
    )
    close_new: pd.Series = (2 * close_normalized) - 1
    smooth: pd.Series = close_new.ewm(span=5, adjust=True).mean()

    return (
        pd.Series((np.log((1 + smooth) / (1 - smooth)))).ewm(span=3, adjust=True).mean()
    )


def min_max_normalization_indicator(close: pd.Series, window: int = 4) -> pd.Series:
    """Indicator based on min max normalization over a specific window.

    Args:
        close (pd.Series): The time series to normalize.
        window (int, optional): The lookback for normalizing the time series. Defaults to 4.

    Returns:
        pd.Series: The normalized time series.
    """
    return (close - close.rolling(window).min()) / (
        close.rolling(window).max() - close.rolling(window).min()
    )


def z_score_indicator(close: pd.Series, window: int = 14) -> pd.Series:
    """Compute the z-score of a time series over a window.

    Args:
        close (pd.Series): The time series to compute the z-score on.
        window (int, optional): The window length of the data used to compute the z-score. Defaults to 14.

    Returns:
        pd.Series: The z-score of the time series.
    """
    return (close - close.rolling(window=window).mean()) / close.rolling(
        window=window
    ).std()


def mean_break_out_indicator(close: pd.Series, window: int = 20) -> pd.Series:
    """Compute the MBO, mean breakout indicator on a time series. It compares the difference between the closing price of a candle and a moving average over N periods to the difference between the min and max value of the closing price over the same N periods.

    Args:
        close (pd.Series): The time series to process.
        window (int, optional): The window length of the data used to compute the MBO. Defaults to 20.

    Returns:
        pd.Series: The MBO indicator.
    """
    return (close - close.ewm(window).mean()) / (
        close.rolling(window).max() - close.rolling(window).min()
    )


def trend_intensity_indicator(close: pd.Series, window: int = 20) -> pd.Series:
    """The Trend Intensity indicator is a measure of the accumulation of the number of periods the price is above the EMA against the price is under the EMA. Given a number of periods equal to the number of periods of the EMA it gives an information on the trend strength and if the trend will tend to reverse.

    Args:
        close (pd.Series): The time series to process.
        window (int, optional): The window length of the data used to compute the trend intensity. Defaults to 20.

    Returns:
        pd.Series: The trend intensity indicator.
    """
    return (
        100
        * pd.DataFrame({"EMA": close.ewm(window).mean().values, "Close": close.values})
        .apply(lambda row: 1 if row["Close"] > row["EMA"] else 0, axis=1)
        .rolling(window)
        .sum()
        / window
    )


def ppo_slope_indicator(close: pd.Series, slope_window: int = 4) -> pd.Series:
    ppo_histo = ppo_hist(close=close)

    def calculate_slope(ppo_values: pd.Series) -> float:
        x = np.arange(1, len(ppo_values) + 1, 1)
        y = np.array(ppo_values)
        m, c = np.polyfit(x, y, 1)
        return m

    return ppo_histo.rolling(slope_window).apply(calculate_slope)


def detect_windowing_support_resistance(high: pd.Series, low: pd.Series) -> list[float]:
    """Detects the supports and resistances according to the windowing method

    Args:
        high (pd.Series): The high series from the OHLCV Data.
        low (pd.Series): The low series from the OHLCV Data.

    Returns:
        list[float]: The supports and resistances levels.
    """
    pivots: list[float] = []
    max_list = []
    min_list = []
    for i in range(5, len(high) - 5):
        # taking a window of 9 candles
        high_range = high.iloc[i - 5 : i + 4]
        current_max = high_range.max()
        # if we find a new maximum value, empty the max_list
        if current_max not in max_list:
            max_list = []
        max_list.append(current_max)
        # if the maximum value remains the same after shifting 5 times
        if len(max_list) == 5 and is_far_from_level(
            current_max,
            pivots,
            high,
            low,
        ):
            pivots.append(float(current_max))

        low_range = low[i - 5 : i + 5]
        current_min = low_range.min()
        if current_min not in min_list:
            min_list = []
        min_list.append(current_min)
        if len(min_list) == 5 and is_far_from_level(
            current_min,
            pivots,
            high,
            low,
        ):
            pivots.append(float(current_min))
    return pivots


def detect_fractal_support_resistance(high: pd.Series, low: pd.Series) -> list[float]:
    """Detects the supports and resistances according to the fractal method.

    Args:
        high (pd.Series): The high series from the OHLCV Data.
        low (pd.Series): The low series from the OHLCV Data.

    Returns:
        list[float]: The supports and resistances levels.
    """

    def is_fractal_support(low: pd.Series, i: int) -> bool:
        """Determine bullish fractal supports.

        Args:
            low (pd.Series): The low series from the OHLCV Data.
            i (int): The current index.

        Returns:
            bool: Whether it's a support or not according to fractal method.
        """
        return (
            low[i] < low[i - 1]
            and low[i] < low[i + 1]
            and low[i + 1] < low[i + 2]
            and low[i - 1] < low[i - 2]
        )

    def is_fractal_resistance(high: pd.Series, i: int) -> bool:
        """Determine bearish fractal resistances.

        Args:
            low (pd.Series): The high series from the OHLCV Data.
            i (int): The current index.

        Returns:
            bool: Whether it's a resistance or not according to fractal method.
        """
        return (
            high[i] > high[i - 1]
            and high[i] > high[i + 1]
            and high[i + 1] > high[i + 2]
            and high[i - 1] > high[i - 2]
        )

    levels: list[float] = []
    k = 5
    for i in range(k, high.shape[0] - k):
        if is_fractal_support(low, i):
            if is_far_from_level(
                low.iloc[i],
                levels,
                high,
                low,
            ):
                levels.append(float(low.iloc[i]))
        elif is_fractal_resistance(high, i):
            if is_far_from_level(
                high.iloc[i],
                levels,
                high,
                low,
            ):
                levels.append(float(high.iloc[i]))

    return levels


def detect_profiled_support_resistance(
    close: pd.Series,
    volume: Optional[pd.Series] = None,
    kde_factor: float = 0.075,
    total_levels: str | int = "all",
) -> pd.DataFrame:
    """Detect the support and resistance level over an historical time period using price and volume.

    Args:
        close (pd.Series[float | int]): The price history usually closing price.
        volume (pd.Series[float | int], optional): The volume history. Defaults to None.
        kde_factor (float, optional):  The coefficient used to calculate the estimator bandwidth. The higher coefficient is the strongest levels will only be detected. Defaults to 0.075.
        total_levels (str | int, optional): The total number of levels to detect. If "all" is provided, all levels will be detected. Defaults to "all".

    Returns:
        pd.DataFrame: The DataFrame containing the levels [min price, max price] and weights [0, 1] associated with each.
    """
    if volume is not None:
        assert len(close) == len(
            volume
        ), "Error, provide same size price and volume Series."

    # Generate a number of sample of the complete price history range in order to apply density estimation.
    xr = np.linspace(start=close.min(), stop=close.max(), num=len(close))

    # Generate the kernel density estimation of the price weighted by volume over a certain number of sample xr.
    # It's possible to interpolate less precisely with decreasing the num parameter above.
    if volume is not None:
        estimated_density = stats.gaussian_kde(
            dataset=close, weights=volume, bw_method=kde_factor
        )(xr)
    else:
        estimated_density = stats.gaussian_kde(dataset=close, bw_method=kde_factor)(xr)

    def min_max_scaling(
        to_scale_array: npt.NDArray[np.float64],
        min_limit: int = 0,
        max_limit: int = 1,
    ) -> npt.NDArray[np.float64]:
        """Min max scaling between 0 and 1.

        Args:
            to_scale_array (npt.NDArray[np.float64]): The array to scale.
            min_limit (int, optional): The lower limit of the range. Defaults to 0.
            max_limit (int, optional): The higher limit of the range. Defaults to 1.

        Returns:
            npt.NDArray[np.float64]: The scaled array.
        """
        return (to_scale_array - to_scale_array.min(axis=0)) / (
            to_scale_array.max(axis=0) - to_scale_array.min(axis=0)
        ) * (max_limit - min_limit) + min_limit

    # Find the index of the peaks over on a signal, here the estimated density.
    peaks, _ = signal.find_peaks(estimated_density)

    levels = xr[peaks]
    weights = min_max_scaling(estimated_density[peaks])

    df = pd.DataFrame({"levels": levels, "weights": weights})

    if isinstance(total_levels, int):
        assert (
            total_levels > 0
        ), "Error, provide a positive not null value for the total_levels parameter."

        if total_levels < len(levels):
            return df.sort_values(by="weights", ascending=False).head(total_levels)

    return df


def is_far_from_level(
    value: float, levels: list[float], high: pd.Series, low: pd.Series
) -> bool:
    """Function to make sure the new level area does not exist already

    Args:
        value (float): The value to check with the existing levels.
        levels (list[float]): The list of levels.
        high (pd.Series): The high series from the OHLCV Data.
        low (pd.Series): The low series from the OHLCV Data.

    Returns:
        bool: Whether the level is far from the others or not.
    """
    return np.sum([abs(value - level) < np.mean(high - low) for level in levels]) == 0



import numpy as np
import numpy.typing as npt
import pandas as pd
from typing import Callable
from abc import abstractmethod
from scipy.optimize import minimize
from quant_invest_lab.backtest import ohlc_long_only_backtester


class ConditionOptimizer:
    def __init__(
        self,
        long_entry_function: Callable[[pd.Series, pd.Series], bool],
        long_exit_function: Callable[[pd.Series, pd.Series, int], bool],
    ) -> None:
        self._long_exit_function: Callable[
            [pd.Series, pd.Series, int], bool
        ] = long_exit_function
        self._long_entry_function: Callable[
            [pd.Series, pd.Series], bool
        ] = long_entry_function

    def _check_range_value(self, value_range: tuple[float | int, float | int]) -> None:
        assert (
            len(value_range) == 2
        ), "range must be a integer or floating tuple of length 2"
        assert value_range[0] < value_range[1], "range must be ascending"
        assert (
            value_range[0] > 0
            and value_range[0] < 1
            and value_range[1] > 0
            and value_range[1] < 1
        ), "range must be positive and between 0 and 1 as a percentage."

    @abstractmethod
    def _compute_return():
        pass

    @abstractmethod
    def optimize():
        pass


class TakeProfitOptimizer(ConditionOptimizer):
    def _compute_return(
        self,
        take_profit: npt.NDArray[np.float64],
        args: list[pd.DataFrame],
    ) -> float:
        return -ohlc_long_only_backtester(
            args[0],
            long_entry_function=self._long_entry_function,
            long_exit_function=self._long_exit_function,
            take_profit=take_profit[0],
            parameter_optimization=True,
        )  # type: ignore

    def optimize(
        self,
        dataframe: pd.DataFrame,
        take_profit_bounds: tuple[float | int, float | int] = (
            0.001,
            0.3,
        ),
    ) -> tuple[float | int, float | int]:
        """Optimize take profit values using `scipy.optimize.minimize` and a SLSQP method.

        Args:
        --------
            dataframe (pd.DataFrame): The dataframe containing the OHLCV data to optimize on the strategy with a given take profit.
            take_profit_bounds (tuple[float  |  int, float  |  int], optional): The tuple of boundaries for the TP between 0.001 and 1. Defaults to (0.001, 0.3)

        Returns:
        --------
            tuple[float | int, float | int]: The result from the optimization process under the form of a tuple containing the optimized take profile values.
        """
        self._check_range_value(take_profit_bounds)

        results = [
            minimize(
                self._compute_return,
                x0=[guess],
                method="SLSQP",  # "SLSQP", #L-BFGS-B
                bounds=tuple(
                    [
                        take_profit_bounds,
                    ]
                ),
                args=[dataframe],
            )
            for guess in np.linspace(
                take_profit_bounds[0],
                take_profit_bounds[-1],
                10,  # Number of init guess to generate
            )
        ]

        results.sort(key=lambda x: x.fun)
        opt_results = results[0]
        return opt_results.x[0]


class StopLossOptimizer(ConditionOptimizer):
    def _compute_return(
        self,
        stop_loss: npt.NDArray[np.float64],
        args: list[pd.DataFrame],
    ) -> float:
        return -ohlc_long_only_backtester(
            args[0],
            long_entry_function=self._long_entry_function,
            long_exit_function=self._long_exit_function,
            stop_loss=stop_loss[0],
            parameter_optimization=True,
        )  # type: ignore

    def optimize(
        self,
        dataframe: pd.DataFrame,
        stop_loss_bounds: tuple[float | int, float | int] = (
            0.001,
            0.3,
        ),
    ) -> tuple[float | int, float | int]:
        """Optimize stop loss values using `scipy.optimize.minimize` and a SLSQP method.

        Args:
        --------
            dataframe (pd.DataFrame): The dataframe containing the OHLCV data to optimize on the strategy with a given stop loss.
            take_profit_and_stop_loss_bounds (tuple[float  |  int, float  |  int], optional): The tuple of boundaries for the SL between 0.001 and 1. Defaults to (0.001, 0.3)

        Returns:
        --------
            tuple[float | int, float | int]: The result from the optimization process under the form of a tuple containing the optimized stop loss values.
        """
        self._check_range_value(stop_loss_bounds)

        results = [
            minimize(
                self._compute_return,
                x0=[guess],
                method="SLSQP",  # "SLSQP", #L-BFGS-B
                bounds=tuple(
                    [
                        stop_loss_bounds,
                    ]
                ),
                args=[dataframe],
            )
            for guess in np.linspace(
                stop_loss_bounds[0],
                stop_loss_bounds[-1],
                5,  # Number of init guess to generate
            )
        ]

        results.sort(key=lambda x: x.fun)
        opt_results = results[0]
        return opt_results.x[0]


class StopLossTakeProfitOptimizer(ConditionOptimizer):
    def _compute_return(
        self,
        take_profit_and_stop_loss: npt.NDArray[np.float64],
        args: list[pd.DataFrame],
    ) -> float:
        return -ohlc_long_only_backtester(
            args[0],
            long_entry_function=self._long_entry_function,
            long_exit_function=self._long_exit_function,
            take_profit=take_profit_and_stop_loss[0],
            stop_loss=take_profit_and_stop_loss[-1],
            parameter_optimization=True,
        )  # type: ignore

    def optimize(
        self,
        dataframe: pd.DataFrame,
        take_profit_and_stop_loss_bounds: tuple[float | int, float | int] = (
            0.001,
            0.3,
        ),
    ) -> tuple[float | int, float | int]:
        """Optimize take profit and stop loss values using `scipy.optimize.minimize` and a SLSQP method.

        Args:
        --------
            dataframe (pd.DataFrame): The dataframe containing the OHLCV data to optimize on the strategy with a given take profit and stop loss.
            take_profit_and_stop_loss_bounds (tuple[float  |  int, float  |  int], optional): The tuple of boundaries for the TP and SL between 0.001 and 1. Defaults to (0.001, 0.3)

        Returns:
        --------
            tuple[float | int, float | int]: The result from the optimization process under the form of a tuple containing the optimized (in this order) take profit and stop loss values.
        """
        self._check_range_value(take_profit_and_stop_loss_bounds)

        results = [
            minimize(
                self._compute_return,
                x0=[guess, guess],
                method="SLSQP",  # "SLSQP", #L-BFGS-B
                bounds=tuple(
                    [
                        take_profit_and_stop_loss_bounds,
                        take_profit_and_stop_loss_bounds,
                    ]
                ),
                args=[dataframe],
            )
            for guess in np.linspace(
                take_profit_and_stop_loss_bounds[0],
                take_profit_and_stop_loss_bounds[-1],
                10,  # Number of init guess to generate
            )
        ]

        results.sort(key=lambda x: x.fun)
        opt_results = results[0]
        return opt_results.x[0], opt_results.x[1]
